#!/usr/bin/env python3

"""Command-line interface for genetic algorithm + neural network generational propagation."""

# Standard library
import sys
import click

# 3rd party packages

# Local source
from parametrization_clean.app import run_application


@click.command('cli', short_help="Run one generation of GA.")
@click.option('-g', '--generation_number', default=1, show_default=True, type=click.IntRange(min=1), required=True,
              help="Generation number for the genetic algorithm. Default = 1 = first generation.")
@click.option('-t', '--training_path', type=click.Path(), required=True,
              help="File path with reference/training set files.")
@click.option('-p', '--population_path', type=click.Path(), required=True,
              help="File path for generational genetic algorithm output.")
@click.option('-c', '--config_path', type=click.Path(),
              help="File path with location of (optional) user configuration file.")
def main(generation_number, training_path, population_path, config_path):
    """Command-line interface for genetic algorithm + neural network generational propagation.
    Runs one generation of the GA.
    """
    # TODO: Make population size a command line option instead of a user configuration file option?
    click.echo("Generation Number: {}".format(generation_number))
    click.echo("Retrieving reference data from: {}".format(training_path))
    click.echo("Retrieving user configuration from: {}".format(config_path if config_path
                                                               else "no user config provided...using DEFAULT config."))
    click.echo("Outputting generational genetic algorithm data to: {}".format(population_path))

    response = run_application(generation_number, training_path, population_path, config_path)

    click.echo("{}: {}".format(response.type, response.message))
    return response.status_code


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



#!/usr/bin/env python

"""
Module that contains interface for adaptive genetic algorithm methods to be used for an Individual.
New adaptation strategies can be added as classes, so long as they implement the abstraction presented here.

__author__ = "Chad Daksha"
"""

# Standard library
import abc
from typing import List, Tuple

# 3rd party packages

# Local source


class IAdaptationStrategy(metaclass=abc.ABCMeta):

    @staticmethod
    @abc.abstractmethod
    def adaptation(average_cost: float, minimum_cost: float, parent_costs: Tuple[float, float], **kwargs) \
            -> Tuple[float, List[float]]:
        raise NotImplementedError




# Standard library

# 3rd party packages

# Local source
from parametrization_clean.domain.adaptation.factory import AdaptationFactory
from parametrization_clean.domain.adaptation.xiao import XiaoAdapt
from parametrization_clean.domain.adaptation.srinivas import SrinivasAdapt


def test_get_xiao_adaptation():
    xiao_adapt = AdaptationFactory.create_executor('xiao')
    assert xiao_adapt == XiaoAdapt


def test_get_srinivas_adaptation():
    srinivas_adapt = AdaptationFactory.create_executor('srinivas')
    assert srinivas_adapt == SrinivasAdapt





# Standard library
import os

# 3rd party packages
import pytest

# Local source
from parametrization_clean.infrastructure.utils.reax_reader import ReaxReader, write_ffield, read_array
from parametrization_clean.domain.utils.helpers import get_param


@pytest.fixture(autouse=True)
@pytest.mark.usefixtures("reax_output_dir_path")
def module_setup_teardown(reax_output_dir_path):
    # Setup
    yield
    # Teardown
    if os.path.isfile(os.path.join(reax_output_dir_path, 'ffield')):
        os.remove(os.path.join(reax_output_dir_path, 'ffield'))


@pytest.fixture()
@pytest.mark.usefixtures("training_set_dir_path")
def reax_io_obj(training_set_dir_path):
    return ReaxReader(training_set_dir_path)


@pytest.mark.usefixtures("training_set_dir_path")
def test_reax_io_init(reax_io_obj, training_set_dir_path):
    assert reax_io_obj.dir_path == training_set_dir_path


def test_reax_io_read_ffield(reax_io_obj):
    ffield, atom_types = reax_io_obj.read_ffield()
    assert isinstance(ffield, dict)
    assert isinstance(atom_types, dict)
    assert get_param([2, 14, 1], ffield) == 1.8862
    assert get_param([2, 14, 25], ffield) == -3.0614
    assert get_param([3, 38, 4], ffield) == -0.6944
    assert get_param([4, 19, 1], ffield) == 0.0987
    assert get_param([5, 94, 5], ffield) == 2.0000
    assert atom_types[3][37 - 1][0] == '3'
    assert atom_types[3][37 - 1][1] == '14'


def test_reax_io_read_fort99(reax_io_obj):
    fort99_results = reax_io_obj.read_fort99()
    assert fort99_results[0][1] == 1.0950
    assert fort99_results[62][1] == 107.2990
    assert fort99_results[65][2] == 0.0500
    assert fort99_results[-1][0] == -0.7802


def test_reax_io_read_params(reax_io_obj):
    param_keys, param_increments, param_bounds = reax_io_obj.read_params()
    assert param_keys[0] == [2, 14, 1]
    assert param_increments[0] == 0.001
    assert param_bounds[0] == [1.8, 2.5]
    assert param_keys[22] == [4, 18, 1]
    assert param_increments[22] == 0.010
    assert param_bounds[22] == [0.02, 0.30]
    assert param_keys[-1] == [5, 94, 7]
    assert param_increments[-1] == 0.050
    assert param_bounds[-1] == [1.1, 4.0]


@pytest.mark.usefixtures("training_set_dir_path", "reax_output_dir_path")
def test_write_ffield(reax_io_obj, training_set_dir_path, reax_output_dir_path):
    input_path = os.path.join(training_set_dir_path, "ffield")
    output_path = os.path.join(reax_output_dir_path, "ffield")
    ffield, atom_types = reax_io_obj.read_ffield()
    write_ffield(input_path, output_path, ffield, atom_types)
    new_reax_io_obj = ReaxReader(reax_output_dir_path)
    new_ffield, new_atom_types = new_reax_io_obj.read_ffield()
    assert get_param([2, 14, 1], new_ffield) == 1.8862
    assert get_param([2, 14, 25], new_ffield) == -3.0614
    assert get_param([3, 38, 4], new_ffield) == -0.6944
    assert get_param([4, 19, 1], new_ffield) == 0.0987
    assert get_param([5, 94, 5], new_ffield) == 2.0000
    assert new_atom_types[3][37 - 1][0] == '3'
    assert new_atom_types[3][37 - 1][1] == '14'



#!/usr/bin/env python

"""Uses ANN to propagate population several times, i.e., a nested GA propagator that runs for several iterations
without performing any ReaxFF optimizations. Requires TensorFlow 2.0.
"""

# Standard library
from typing import List

# 3rd party packages

# Local source
from parametrization_clean.domain.individual import Individual
from parametrization_clean.domain.neural_network.ann import FeedForwardNet
from parametrization_clean.use_case.port.settings_repository import IAllSettings
from parametrization_clean.use_case.port.population_repository import IPopulationRepository
from parametrization_clean.use_case.population_propagator import PopulationPropagator


class GeneticNeuralNetPropagator:

    def __init__(self, settings_repository: IAllSettings, population_repository: IPopulationRepository):
        self.neural_net_settings = settings_repository.neural_net_settings
        self.population_repository = population_repository

        training_population = population_repository.get_previous_n_populations(
            self.neural_net_settings.num_populations_to_train_on)
        self.neural_net = FeedForwardNet(training_population,
                                         self.neural_net_settings.verbosity,
                                         self.neural_net_settings.train_fraction,
                                         self.neural_net_settings.num_epochs)

        self.population_propagator = PopulationPropagator(settings_repository, population_repository)
        self.root_individual = self.population_repository.get_root_individual()
        self.error_strategy = settings_repository.strategy_settings.error_strategy
        self.population_size = settings_repository.ga_settings.population_size

    def train_neural_net(self):
        model, history = self.neural_net.execute()
        return model, history

    def execute(self, parents: List[Individual]):
        model, history = self.train_neural_net()
        if self.final_ann_accuracy_is_poor(history):
            final_generation = self.run_without_ann(parents, model)
        else:
            final_generation = self.run_with_ann(parents, model)
        return final_generation, model, history

    def run_without_ann(self, parents, model):
        final_generation, _ = self.propagate_first(parents, model)
        return final_generation

    def run_with_ann(self, parents, model):
        next_generation, best_master_parents = self.propagate_first(parents, model)
        final_generation = self.propagate_remaining(next_generation, model)
        return final_generation

    def final_ann_accuracy_is_poor(self, history):
        accuracy_is_poor = False
        final_val_r_squared = history.history['val_r_square'][-1]
        if final_val_r_squared < self.neural_net_settings.minimum_validation_r_squared:
            accuracy_is_poor = True
        return accuracy_is_poor

    def propagate_first(self, parents: List[Individual], model):
        """First propagation/iteration in nested genetic algorithm. Preserve best two parents from master GA."""
        next_generation = self.population_propagator.execute(parents)
        # Update costs for all except top 2 individuals
        y_predicted = self.neural_net.predict_outputs(model, next_generation[2:])
        self.update_costs(next_generation[2:], y_predicted)

        best_master_parents = next_generation[0:2]
        return next_generation, best_master_parents

    def propagate_remaining(self, population: List[Individual], model) -> List[Individual]:
        """Run remaining nested genetic algorithm iterations. Preserve best two parents from master GA."""
        next_generation = population
        for i in range(self.neural_net_settings.num_nested_ga_iterations):
            next_generation = self.population_propagator.execute(next_generation)
            y_predicted = self.neural_net.predict_outputs(model, next_generation[2:])
            self.update_costs(next_generation[2:], y_predicted)
        return next_generation

    def update_costs(self, population: List[Individual], y_predicted):
        costs = self.neural_net.compute_costs(y_predicted, self.root_individual, self.error_strategy)
        for individual, y_pred_row, cost in zip(population, y_predicted, costs):
            individual.reax_energies = y_pred_row
            individual.cost = cost




# Standard library

# 3rd party packages

# Local source
from parametrization_clean.domain.cost.factory import ErrorFactory
from parametrization_clean.domain.cost.reax_error import ReaxError


def test_get_reax_error():
    reax_error_calculator = ErrorFactory.create_executor('reax_error')
    assert reax_error_calculator == ReaxError




#!/usr/bin/env python

"""
Module that contains interface for mutation methods to be used for an Individual.
New mutation strategies can be added as classes, so long as they implement the abstraction presented here.

__author__ = "Chad Daksha"
"""

# Standard library
import abc

# 3rd party packages


# Local source
from parametrization_clean.domain.individual import Individual
from parametrization_clean.domain.root_individual import RootIndividual


class IMutationStrategy(metaclass=abc.ABCMeta):

    @staticmethod
    @abc.abstractmethod
    def mutation(parent: Individual, root_individual: RootIndividual, **kwargs) -> Individual:
        raise NotImplementedError




# Standard library
import os
from pathlib import Path
import shutil

# 3rd party packages
import pytest

# Local source
from parametrization_clean.infrastructure.repository.from_files import PopulationFileRepository
from parametrization_clean.domain.root_individual import RootIndividual, FirstGenerationRootIndividual
from parametrization_clean.infrastructure.utils.response_object import (ResponseSuccess,
                                                  ResponseWarning,
                                                  ResponseFailure)
from tests.use_case.test_population_propagator import all_settings
from tests.infrastructure.utils.test_reaxff_reader import reax_io_obj


@pytest.fixture(autouse=True)
@pytest.mark.usefixtures("reax_output_dir_path")
def module_setup_teardown(reax_output_dir_path):
    # Setup
    yield
    # Teardown
    if os.path.isdir(os.path.join(reax_output_dir_path, 'generation-3')):
        shutil.rmtree(os.path.join(reax_output_dir_path, 'generation-3'))


@pytest.fixture()
@pytest.mark.usefixtures("training_set_dir_path", "reax_output_dir_path")
def file_repository(all_settings, training_set_dir_path, reax_output_dir_path):
    return PopulationFileRepository(training_set_dir_path, reax_output_dir_path, all_settings,
                                    current_generation_number=3)


@pytest.fixture()
@pytest.mark.usefixtures("training_set_dir_path", "reax_output_dir_path")
def first_generation_file_repository(all_settings, training_set_dir_path, reax_output_dir_path):
    return PopulationFileRepository(training_set_dir_path, reax_output_dir_path, all_settings,
                                    current_generation_number=1)


@pytest.fixture()
@pytest.mark.usefixtures("training_set_dir_path", "reax_output_dir_path")
def second_generation_file_repository(all_settings, training_set_dir_path, reax_output_dir_path):
    return PopulationFileRepository(training_set_dir_path, reax_output_dir_path, all_settings,
                                    current_generation_number=2)


@pytest.mark.usefixtures("training_set_dir_path", "reax_output_dir_path")
def test_file_repository_init(file_repository, all_settings, training_set_dir_path, reax_output_dir_path):
    assert file_repository.training_set_path == Path(training_set_dir_path)
    assert file_repository.population_path == Path(reax_output_dir_path)
    assert file_repository.current_generation_number == 3
    assert file_repository.population_size == 4

    file_repository = PopulationFileRepository(training_set_dir_path,
                                               os.path.join(reax_output_dir_path, 'test_output_directory'),
                                               all_settings, current_generation_number=3)
    assert file_repository.training_set_path == Path(training_set_dir_path)
    assert file_repository.population_path == Path(os.path.join(reax_output_dir_path, 'test_output_directory'))
    assert file_repository.current_generation_number == 3
    assert file_repository.population_size == 4
    assert os.path.isdir(os.path.join(reax_output_dir_path, 'test_output_directory'))
    os.rmdir(os.path.join(reax_output_dir_path, 'test_output_directory'))


def test_get_root_individual_first_generation(reax_io_obj, first_generation_file_repository):
    root_individual = first_generation_file_repository.get_root_individual()

    ffield, _ = reax_io_obj.read_ffield()
    param_keys, _, _ = reax_io_obj.read_params()

    assert isinstance(root_individual, FirstGenerationRootIndividual)
    assert isinstance(root_individual.root_ffield, dict)
    assert root_individual.param_keys == param_keys


def test_get_root_individual_second_generation(reax_io_obj, second_generation_file_repository):
    if os.path.isdir(second_generation_file_repository.reference_path):
        shutil.rmtree(second_generation_file_repository.reference_path)

    root_individual = second_generation_file_repository.get_root_individual()

    ffield, _ = reax_io_obj.read_ffield()
    param_keys, _, _ = reax_io_obj.read_params()

    assert isinstance(root_individual, RootIndividual)
    assert isinstance(root_individual.root_ffield, dict)
    assert root_individual.param_keys == param_keys
    assert os.path.isdir(second_generation_file_repository.reference_path)
    assert os.path.exists(os.path.join(second_generation_file_repository.reference_path, "fort.99"))


def test_get_root_individual(reax_io_obj, file_repository):
    root_individual = file_repository.get_root_individual()

    ffield, _ = reax_io_obj.read_ffield()
    fort99_data = reax_io_obj.read_fort99()
    param_keys, _, _ = reax_io_obj.read_params()

    dft_energies = [row[1] for row in fort99_data]
    weights = [row[2] for row in fort99_data]

    assert os.path.isdir(file_repository.reference_path)
    assert os.path.exists(os.path.join(file_repository.reference_path, "fort.99"))

    assert isinstance(root_individual, RootIndividual)
    assert isinstance(root_individual.root_ffield, dict)
    assert list(root_individual.dft_energies) == dft_energies
    assert list(root_individual.weights) == weights
    assert root_individual.param_keys == param_keys


def test_get_population(file_repository):
    first_generation_population, successfully_retrieved_case_numbers = \
        file_repository.get_population(generation_number=1)
    assert len(first_generation_population) == 4
    assert len(successfully_retrieved_case_numbers) == 4
    assert first_generation_population[0].cost == pytest.approx(7253274.4676, rel=100)
    assert first_generation_population[1].cost == pytest.approx(4094926.0857, rel=100)
    assert isinstance(first_generation_population[2].cost, float)
    assert first_generation_population[-1].cost == pytest.approx(248741.8623, rel=100)

    file_repository.population_size = 5
    first_generation_population, _ = file_repository.get_population(generation_number=1)
    assert len(first_generation_population) == 4
    assert first_generation_population[0].cost == pytest.approx(7253274.4676, rel=100)
    assert first_generation_population[1].cost == pytest.approx(4094926.0857, rel=100)
    assert isinstance(first_generation_population[2].cost, float)
    assert first_generation_population[-1].cost == pytest.approx(248741.8623, rel=100)


def test_read_population_range(file_repository):
    first_and_second_generation_population = file_repository.read_population_range(lower_bound=1,
                                                                                   upper_bound=3)
    assert len(first_and_second_generation_population) == 8
    assert first_and_second_generation_population[0].cost == pytest.approx(7253274.4676, rel=100)
    assert first_and_second_generation_population[1].cost == pytest.approx(4094926.0857, rel=100)
    assert isinstance(first_and_second_generation_population[2].cost, float)
    assert first_and_second_generation_population[3].cost == pytest.approx(248741.8623, rel=100)
    assert first_and_second_generation_population[4].cost == pytest.approx(644975.3705, rel=100)
    assert first_and_second_generation_population[5].cost == pytest.approx(710538.2521, rel=100)
    assert isinstance(first_and_second_generation_population[6].cost, float)
    assert isinstance(first_and_second_generation_population[7].cost, float)


def test_get_previous_n_populations(file_repository):
    previous_two_populations = file_repository.get_previous_n_populations(num_populations=2)

    assert len(previous_two_populations) == 8
    assert previous_two_populations[0].cost == pytest.approx(7253274.4676, rel=100)
    assert previous_two_populations[1].cost == pytest.approx(4094926.0857, rel=100)
    assert isinstance(previous_two_populations[2].cost, float)
    assert previous_two_populations[3].cost == pytest.approx(248741.8623, rel=100)
    assert previous_two_populations[4].cost == pytest.approx(644975.3705, rel=100)
    assert previous_two_populations[5].cost == pytest.approx(710538.2521, rel=100)
    assert isinstance(previous_two_populations[6].cost, float)
    assert isinstance(previous_two_populations[7].cost, float)

    previous_one_population = file_repository.get_previous_n_populations(num_populations=1)
    assert len(previous_one_population) == 4
    assert previous_two_populations[0].cost == pytest.approx(644975.3705, rel=100)
    assert previous_two_populations[1].cost == pytest.approx(710538.2521, rel=100)
    assert isinstance(previous_two_populations[2].cost, float)
    assert isinstance(previous_two_populations[3].cost, float)


@pytest.mark.usefixtures('get_individuals')
def test_write_individual(file_repository, reax_output_dir_path):
    population, _ = file_repository.get_population(generation_number=1)
    individual = population[0]

    generation_path = os.path.join(reax_output_dir_path, 'generation-3')
    os.mkdir(generation_path)
    child_dir = os.path.join(generation_path, 'child-0')
    response = file_repository.write_individual(individual, child_dir=child_dir)

    assert isinstance(response, ResponseSuccess)
    assert response.message == "Child successfully written at {}.".format(child_dir)

    response = file_repository.write_individual(individual, child_dir=child_dir)
    assert isinstance(response, ResponseWarning)
    assert response.message == "Files in child directory {} were overwritten.".format(child_dir)

    response = file_repository.write_individual(individual)
    assert isinstance(response, ResponseFailure)
    assert response.message == "Child directory not provided."


def test_write_population(file_repository, reax_output_dir_path):
    population, _ = file_repository.get_population(generation_number=2)

    generation_path = os.path.join(reax_output_dir_path, 'generation-3')
    response = file_repository.write_population(population, generation_number=3)
    assert isinstance(response, ResponseSuccess)
    assert response.message == "Generation successfully written at {}".format(generation_path)

    shutil.rmtree(generation_path)
    os.mkdir(generation_path)
    response = file_repository.write_population(population, generation_number=3)
    assert isinstance(response, ResponseWarning)
    assert response.message == "Population generation directory {} overwritten.".format(generation_path)



#!/usr/bin/env python

"""Module with standard ReaxFF error function implementation."""

# Standard library

# 3rd party packages

# Local source
from parametrization_clean.domain.cost.strategy import IErrorStrategy


class ReaxError(IErrorStrategy):

    @staticmethod
    def error(reax_val, dft_val, weight, **kwargs) -> float:
        """Calculate ReaxFF error using error = ((reax_pred - true_val)/weight)^2."""
        return ((reax_val - dft_val) / weight) ** 2



#!/usr/bin/env python

"""Factory for selection algorithms allowed for usage."""

# Standard library

# 3rd party packages

# Local source
from parametrization_clean.domain.selection.strategy import ISelectionStrategy
from parametrization_clean.domain.selection.tournament import TournamentSelect


class SelectionFactory:
    """Factory class for creating selection algorithm executor - RegistryHolder design pattern.
    Classes that implement ISelectionStrategy can be registered and utilized through this factory's registry.
    """

    REGISTRY = {}
    """Internal registry for available selection methods. Users can specify from one of the
    `algorithm_name` strings available in the dictionary, mapping `algorithm_name` to the corresponding class
    implementing that algorithm.
    For example, "tournament" maps to the Tournament selection algorithm;
    users can specify the `selection_strategy` in the user config.json file to use this algorithm.

    CURRENTLY, ONLY THE TOURNAMENT SELECTION ALGORITHM HAS BEEN IMPLEMENTED AND IS SUPPORTED.
    """

    @classmethod
    def register(cls, algorithm_name: str, selection_class):
        """Register a selection strategy with a string key. Useful for abstraction and dynamic retrieval
        of different algorithms in configuration file. Using this factory, one can easily implement a crossover
        algorithm (ex. MySelectionClass) that follows ISelectionStrategy, then use
        "SelectionFactory.register('my_selection_class_name')"
        to generate a corresponding string reference for that selection strategy.

        Parameters
        ----------
        algorithm_name: str
            Name that one wishes to assign to the designated `selection_class`/algorithm.
        selection_class
            Class that one wishes to associate/register with `algorithm_name`.
        Returns
        -------
        mutation_class
            Same as the `selection_class` input parameter.
        """
        cls.REGISTRY[algorithm_name] = selection_class
        return selection_class

    @classmethod
    def create_executor(cls, algorithm_name: str) -> ISelectionStrategy:
        return cls.REGISTRY[algorithm_name]


SelectionFactory.register('tournament', TournamentSelect)




# Standard library

# 3rd party packages

# Local source
from parametrization_clean.infrastructure.utils.response_object import (ResponseSuccess,
                                                                        ResponseWarning,
                                                                        ResponseFailure)


def test_response_success():
    success_obj = ResponseSuccess()

    assert success_obj
    assert success_obj.type == ResponseSuccess.SUCCESS
    assert success_obj.status_code == 0
    assert not success_obj.message


def test_response_success_with_message():
    message = "Successful response."
    success_obj = ResponseSuccess(message)

    assert success_obj
    assert success_obj.type == ResponseSuccess.SUCCESS
    assert success_obj.status_code == 0
    assert success_obj.message == message


def test_response_warning():
    warning_obj = ResponseWarning()

    assert warning_obj
    assert warning_obj.type == ResponseWarning.WARNING
    assert warning_obj.status_code == 0
    assert not warning_obj.message


def test_response_warning_with_message():
    message = "WARNING."
    warning_obj = ResponseWarning(message)

    assert warning_obj
    assert warning_obj.type == ResponseWarning.WARNING
    assert warning_obj.status_code == 0
    assert warning_obj.message == message


def test_response_failure_init():
    failure_obj = ResponseFailure(type_=ResponseFailure.RESOURCE_ERROR, message="ERROR.")

    assert not failure_obj
    assert failure_obj.type == ResponseFailure.RESOURCE_ERROR
    assert failure_obj.message == "ERROR."
    assert failure_obj.value == {'type': ResponseFailure.RESOURCE_ERROR,
                                 'message': failure_obj.message}
    assert failure_obj.status_code == 1


def test_response_failure_build_from_resource_error():
    failure_obj = ResponseFailure.build_resource_error(message="ERROR.")

    assert not failure_obj
    assert failure_obj.type == ResponseFailure.RESOURCE_ERROR
    assert failure_obj.message == "ERROR."
    assert failure_obj.value == {'type': ResponseFailure.RESOURCE_ERROR,
                                 'message': failure_obj.message}
    assert failure_obj.status_code == 1


def test_response_failure_build_from_system_error():
    failure_obj = ResponseFailure.build_system_error(message="ERROR.")

    assert not failure_obj
    assert failure_obj.type == ResponseFailure.SYSTEM_ERROR
    assert failure_obj.message == "ERROR."
    assert failure_obj.value == {'type': ResponseFailure.SYSTEM_ERROR,
                                 'message': failure_obj.message}
    assert failure_obj.status_code == 1


def test_response_failure_build_from_parameters_error():
    failure_obj = ResponseFailure.build_parameters_error(message="ERROR.")

    assert not failure_obj
    assert failure_obj.type == ResponseFailure.PARAMETERS_ERROR
    assert failure_obj.message == "ERROR."
    assert failure_obj.value == {'type': ResponseFailure.PARAMETERS_ERROR,
                                 'message': failure_obj.message}
    assert failure_obj.status_code == 1


def test_response_failure_format_message():
    failure_obj = ResponseFailure.build_parameters_error(message=ValueError("Incorrect value entered!"))
    assert failure_obj.message == "ValueError: Incorrect value entered!"



#!/usr/bin/env python

# Standard library
import random

# 3rd party packages


# Local source
from parametrization_clean.domain.mutation.strategy import IMutationStrategy
from parametrization_clean.domain.individual import Individual
from parametrization_clean.domain.root_individual import RootIndividual


class NakataMutate(IMutationStrategy):

    @staticmethod
    def mutation(parent: Individual, root_individual: RootIndividual, **kwargs) -> Individual:
        """Mutate Child's `params` using Nakata's methodology.
        new_param = old_param + (scale * rand_num * old_param).

        `scale` can be float or List with len(`scale`) = len(self.params),
        e.g., when using param_increments is desired.
        param_bounds are currently being used as (min, max) conditions for params, if they exist.
        if param is outside param_bounds, param is set using uniform distribution with (min, max) bounds.

        :return: New Individual after mutation.
        """
        scale = kwargs.get('nakata_scale', 0.1)
        low = kwargs.get('nakata_rand_lower', -1.0)
        high = kwargs.get('nakata_rand_higher', 1.0)
        new_params = [param + (scale * random.uniform(low, high) * param) for param in parent.params]
        return Individual(new_params, root_individual=root_individual)



#!/usr/bin/env python

"""Module to read, structure, and clean data from reference ReaxFF files:
ffield, fort.99, params.
"""

# Standard library
from typing import List, Tuple, Dict
import os
import re
import textwrap

# 3rd party packages
import numpy as np

# Local source


class ReaxReader(object):

    def __init__(self, dir_path):
        """

        Parameters
        ----------
        dir_path: str
            Directory path containing ffield, fort.99, params files.
        """
        self.dir_path = dir_path

    def read_ffield(self) -> Tuple[Dict, Dict]:
        """Read ffield file into dictionary, mapping each section to corresponding rows.

        ffield[section number] = [parameters corresponding to that section].
        ASSUMES ffield file name is 'ffield'.

        Returns
        -------
        ffield: Dict[int, List]
            ffield[section number] = [parameters corresponding to section]
        atom_types: Dict[int, List]
            atom_types[section number] = [atom types for each section]
        """
        ffield = {}
        atom_types = {}

        # --- Begin Reading FFIELD File ---
        with open(os.path.join(self.dir_path, 'ffield'), 'r') as in_file:
            # Reading 1st section - General Info
            in_file.readline()
            num_general = int(in_file.readline().split().pop(0))
            ffield[1] = read_array(file=in_file, max_rows=num_general, columns=(0,))

            # Reading 2nd section - Atom Info
            ffield[2] = []
            atom_types[2] = []
            # Skipping lines
            num_atoms = int(in_file.readline().split().pop(0))
            for i in range(3):
                in_file.readline()
            for i in range(num_atoms):
                temp = []
                for j in range(4):
                    temp = temp + in_file.readline().split()
                atom_types[2].append(temp.pop(0))  # First value is atom type; ex: 'C', 'S', 'N', etc.
                ffield[2].append([float(param) for param in temp])  # Conversion from str to float

            # Reading 3rd section - Bond Info
            ffield[3] = []
            atom_types[3] = []
            num_bonds = int(in_file.readline().split().pop(0))
            in_file.readline()
            for i in range(num_bonds):
                temp = []
                for j in range(2):
                    temp = temp + in_file.readline().split()
                atom_types[3].append([temp.pop(0), temp.pop(0)])  # First two values are atom types
                ffield[3].append([float(param) for param in temp])

            # Reading 4th section - Off Diagonal Terms
            num_offdiag = int(in_file.readline().split().pop(0))
            ffield[4] = read_array(file=in_file, max_rows=num_offdiag)
            atom_types[4] = ffield[4][:, [0, 1]]  # First two columns are atom types
            ffield[4] = ffield[4][:, 2:]

            # Reading 5th section - Angular Terms
            num_angles = int(in_file.readline().split().pop(0))
            ffield[5] = read_array(file=in_file, max_rows=num_angles)
            atom_types[5] = ffield[5][:, [0, 1, 2]]  # First three columns are atom types
            ffield[5] = ffield[5][:, 3:]

            # Reading 6th section - Torsion Terms
            num_torsions = int(in_file.readline().split().pop(0))
            ffield[6] = read_array(file=in_file, max_rows=num_torsions)
            atom_types[6] = ffield[6][:, [0, 1, 2, 3]]  # First four columns are atom types
            ffield[6] = ffield[6][:, 4:]

            # Reading 7th section - Hydrogen Bonds
            num_hbonds = int(in_file.readline().split().pop(0))
            ffield[7] = read_array(file=in_file, max_rows=num_hbonds)
            atom_types[7] = ffield[7][:, [0, 1, 2]]  # First three columns are atom types
            ffield[7] = ffield[7][:, 3:]

        return ffield, atom_types

    def read_fort99(self) -> List:
        """Read fort.99 output created from simulation. Store output in list; primarily useful for
        error and total error.

        ASSUMES fort99 file name is 'fort.99'.

        Returns
        -------
        results: List
            List with columns of [ffield value | QM/Lit value | Weight].
        """
        results = []

        with open(os.path.join(self.dir_path, 'fort.99'), 'r') as in_file:
            in_file.readline()
            for line in in_file:
                if not line.strip():
                    # Skip empty lines
                    continue
                else:
                    # Extracting all floating points from the line (that are not right next to strings)
                    # All numbers with %.4f formatting are extracted.
                    numbers = re.findall(r"[-+]?\d*\.\d{4}", line)

                    # First three values are reaxFF value, true value, and weight
                    results.append([float(string) for string in numbers[0:3]])

        return results

    def read_params(self) -> Tuple[List, List, List]:
        """Read PARAMS file into lists containing a map for the reference initial parameters and containing the
        parameter min/max bounds, if specified.

        ASSUMES params file name is 'params'.

        Returns
        -------
        param_keys: List[List[int]]
            list containing 3 columns: [section # | line/row # | parameter #].
        param_increments: List[int]
            list containing appropriate increment values for each parameter in the parameter map.
        param_bounds: List[List[float]]
            list containing minimum & maximum values (if specified) for each parameter in the parameter map.
            list format: [minimum value | maximum value].
        """
        param_keys = []
        param_increments = []
        param_bounds = []

        # --- Begin Reading PARAMS File ---
        with open(os.path.join(self.dir_path, 'params'), 'r') as in_file:
            for line in in_file:
                temp = line.split()
                cleaned = []
                for item in temp:
                    try:
                        test = float(item)
                        # Only append non-string values to cleaned list
                        cleaned.append(test)
                    except ValueError:
                        # Skipping string values
                        pass
                param_increments.append(cleaned.pop(3))  # 4th column (delta) is the parameter increment
                cleaned = [int(item) if item.is_integer() else item for item in cleaned]
                param_keys.append(cleaned[0:3])  # First three columns are parameter map
                # temporary fix for param bounds
                try:
                    # 4th and 5th columns contain min and max parameter bounds (if they exist)
                    param_bounds.append(sorted([item for item in cleaned[3:]]))
                except IndexError:
                    param_bounds.append([])

        return param_keys, param_increments, param_bounds


def write_ffield(parent_path: str, output_path: str, ffield: Dict[int, List], atom_types: Dict[int, List]):
    """Using the formatted ffield at `parent_path`, write the provided `ffield` to `output_path`.

    Parameters
    ----------
    parent_path: str
        File path with reference ffield file.
    output_path: str
        File path to output new ffield file.
    atom_types: Dict[int, List]
        atom_types[section number] = [atom types for each section]
    ffield: Dict[int, List]
        ffield[section number] = [parameters corresponding to section]
    """
    if not os.path.exists(parent_path):
        print("ffield file not found in {}...exiting function".format(parent_path))
        return

    # --- Begin Reading root/input FFIELD file and writing to output file ---
    with open(parent_path, 'r') as in_file, open(output_path, 'w') as out_file:
        # Writing 1st section - General Info
        # Assumes general parameters aren't changed, so writes based on reference ffield set
        out_file.write(in_file.readline())
        num_general_line = in_file.readline()
        out_file.write(num_general_line)
        num_general = int(num_general_line.split().pop(0))
        for _ in range(num_general):
            out_file.write(in_file.readline())

        # Writing 2nd section - Atom Info
        # Skipping lines
        num_atoms_line = in_file.readline()
        out_file.write(num_atoms_line)
        num_atoms = int(num_atoms_line.split().pop(0))
        for _ in range(3):
            out_file.write(in_file.readline())
        for i in range(num_atoms):
            atom_line = in_file.readline()  # Line containing atom...ex 'C', 'H', etc.
            atom = [atom_line.split().pop(0)]
            atom_params = ffield[2][i]  # Parameters corresponding to atom in atom section of ffield dict

            # Formatting parameters for output. 'w' = field width, 'p' = decimal precision
            formatted_params = ["{:>{w}.{p}f}".format(param, w=9, p=4) for param in atom_params]
            atom_output = atom + [" "] + formatted_params
            formatted_output = "".join(atom_output)
            formatted_output = textwrap.wrap(formatted_output, width=75)

            # TODO: Explain complicated formatting procedure!
            for item in formatted_output:
                if item is formatted_output[0]:  # First item - contains atom identifier...ex 'C', 'H', etc.
                    out_file.write(
                        " " + "{:<4}".format(item.split()[0]) + "{:>7.4f}".format(float(item.split()[1])) + "".join(
                            ["{:{w}.{p}f}".format(float(param), w=9, p=4) for param in item.split()[2:]]) + "\n")
                elif float(item.split().pop(0)) < 0:  # First number is negative
                    out_file.write("{:12.4f}".format(float(item.split()[0])) + "".join(
                        ["{:>{w}.{p}f}".format(float(param), w=9, p=4) for param in item.split()[1:]]) + "\n")
                else:  # First number is positive
                    out_file.write("{:12.4f}".format(float(item.split()[0])) + "".join(
                        ["{:>{w}.{p}f}".format(float(param), w=9, p=4) for param in item.split()[1:]]) + "\n")

            # Skipping lines in input file to reach next atom
            for _ in range(3):
                in_file.readline()

        # Writing 3rd section - Bond Info
        num_bonds_line = in_file.readline()
        out_file.write(num_bonds_line)
        num_bonds = int(num_bonds_line.split().pop(0))
        out_file.write(in_file.readline())
        for i in range(num_bonds):
            # Combining atom IDs of bonded atoms...ex '1' and '2', etc.
            # with bond parameters corresponding to bonded atoms
            bond_params = [int(atomID) for atomID in atom_types[3][i]] + ffield[3][i]

            # Formatting parameters for output. 'w' = field width, 'p' = decimal precision
            # formatted_params = ["{:>{w}.{p}f}".format(param, w=9, p=4) for param in bond_params]
            formatted_params = []
            for param in bond_params:
                if param is bond_params[0] or param is bond_params[1]:
                    formatted_params.append("{:>{w}d}".format(param, w=3))
                else:
                    formatted_params.append("{:>{w}.{p}f}".format(param, w=9, p=4))
            formatted_output = "".join(formatted_params)
            formatted_output = textwrap.wrap(formatted_output, width=78)

            # TODO: Explain complicated formatting procedure!
            for item in formatted_output:
                if item is formatted_output[0]:  # First item - contains bonded atom pair...ex '1' and '2', etc.
                    out_file.write(item + "\n")
                else:
                    out_file.write(
                        "{:>15.4f}".format(float(item.split()[0]))
                        + "".join(["{:{w}.{p}f}".format(float(param), w=9, p=4) for param in item.split()[1:]])
                        + "\n")

            # Skipping lines in input file to reach next bond term
            for _ in range(2):
                in_file.readline()

        # Writing 4th section - Off Diagonal Terms
        num_offdiag_line = in_file.readline()
        out_file.write(num_offdiag_line)
        num_offdiag = int(num_offdiag_line.split().pop(0))
        # Rejoining first two atom type columns
        offdiag_params = np.concatenate((atom_types[4], ffield[4]), axis=1)
        for i in range(num_offdiag):
            temp_params = offdiag_params[i]

            # Formatting parameters for output. 'w' = field width, 'p' = decimal precision
            formatted_params = []
            for index, param in enumerate(temp_params):
                if index in {0, 1}:
                    formatted_params.append("{:>{w}d}".format(int(param), w=3))
                else:
                    formatted_params.append("{:>{w}.{p}f}".format(param, w=9, p=4))
            formatted_output = "".join(formatted_params)
            formatted_output = textwrap.wrap(formatted_output, width=60)

            # TODO: Explain complicated formatting procedure!
            for item in formatted_output:
                out_file.write(item + "\n")

            # Skipping lines in input file to reach next off diagonal term
            in_file.readline()

        # Writing 5th section - Angular Terms
        num_angles_line = in_file.readline()
        out_file.write(num_angles_line)
        num_angles = int(num_angles_line.split().pop(0))
        # Rejoining first three atom type columns
        angle_params = np.concatenate((atom_types[5], ffield[5]), axis=1)
        for i in range(num_angles):
            temp_params = angle_params[i]

            # Formatting parameters for output. 'w' = field width, 'p' = decimal precision
            formatted_params = []
            for param in temp_params[0:3]:
                formatted_params.append("{:>{w}d}".format(int(param), w=3))
            for param in temp_params[3:]:
                formatted_params.append("{:>{w}.{p}f}".format(param, w=9, p=4))
            formatted_output = "".join(formatted_params)
            formatted_output = textwrap.wrap(formatted_output, width=72)

            # TODO: Explain complicated formatting procedure!
            for item in formatted_output:
                out_file.write(item + "\n")

            # Skipping lines in input file to reach next angle term
            in_file.readline()

        # Writing 6th section - Torsion Terms
        num_torsions_line = in_file.readline()
        out_file.write(num_torsions_line)
        num_torsions = int(num_torsions_line.split().pop(0))
        # Rejoining first four atom type columns
        torsion_params = np.concatenate((atom_types[6], ffield[6]), axis=1)
        for i in range(num_torsions):
            temp_params = torsion_params[i]

            # Formatting parameters for output. 'w' = field width, 'p' = decimal precision
            formatted_params = []
            for param in temp_params[0:4]:
                formatted_params.append("{:>{w}d}".format(int(param), w=3))
            for param in temp_params[4:]:
                formatted_params.append("{:>{w}.{p}f}".format(param, w=9, p=4))
            formatted_output = "".join(formatted_params)
            formatted_output = textwrap.wrap(formatted_output, width=75)

            # TODO: Explain complicated formatting procedure!
            for item in formatted_output:
                out_file.write(item + "\n")

            # Skipping lines in input file to reach next angle term
            in_file.readline()

        # Writing 7th section - Hydrogen Bonds
        num_hbonds_line = in_file.readline()
        out_file.write(num_hbonds_line)
        num_hbonds = int(num_hbonds_line.split().pop(0))
        # Rejoining first three atom type columns
        hbonds_params = np.concatenate((atom_types[7], ffield[7]), axis=1)
        for i in range(num_hbonds):
            temp_params = hbonds_params[i]

            # Formatting parameters for output. 'w' = field width, 'p' = decimal precision
            formatted_params = []
            for param in temp_params[0:3]:
                formatted_params.append("{:>{w}d}".format(int(param), w=3))
            for param in temp_params[3:]:
                formatted_params.append("{:>{w}.{p}f}".format(param, w=9, p=4))
            formatted_output = "".join(formatted_params)
            formatted_output = textwrap.wrap(formatted_output, width=45)

            # TODO: Explain complicated formatting procedure!
            for item in formatted_output:
                out_file.write(item + "\n")

            # Skipping lines in input file to reach next angle term
            in_file.readline()


def read_array(file, skip_header: int = 0, max_rows: int = None, columns: Tuple[int] = None, delimiter=None)\
        -> np.ndarray:
    """Read a specified portion of a file into an array using NumPy's genfromtxt function.

    Parameters
    ----------
    file
        string file path OR FileIO object from which the data is extracted
    skip_header : int
        Number of header lines to skip
    max_rows : int
        Maximum number of rows to read into array
    columns : tuple, default=None
        Tuple of columns from data to read; ex (0, 2, 4) reads cols 0, 2, 4
    delimiter : string, default=None
        Data delimiter; assumes spaces as default.

    Returns
    -------
    np.ndarray
        Numpy array with columns of data
    """
    return np.genfromtxt(file, skip_header=skip_header, max_rows=max_rows, usecols=columns, delimiter=delimiter)



#!/usr/bin/env python

"""Combines domain logic to propagate (generational) genetic algorithm. To be used if the generation number
is greater than one. Applies selection, crossover, mutation, and adaptation operators on the parents (individuals
from the previous generation) to generate better offspring/children (individuals for the next generation).
"""

# Standard library
from typing import List, Tuple
import random
import statistics

# 3rd party packages

# Local source
from parametrization_clean.domain.individual import Individual
from parametrization_clean.use_case.port.settings_repository import IAllSettings
from parametrization_clean.use_case.port.population_repository import IPopulationRepository


class PopulationPropagator:

    def __init__(self, settings_repository: IAllSettings, population_repository: IPopulationRepository):
        self.ga_settings = settings_repository.ga_settings

        self.crossover_rate = self.ga_settings.crossover_rate
        self.mutation_rates = [self.ga_settings.mutation_rate, self.ga_settings.mutation_rate]

        self.mutation_strategy = settings_repository.strategy_settings.mutation_strategy
        self.crossover_strategy = settings_repository.strategy_settings.crossover_strategy
        self.selection_strategy = settings_repository.strategy_settings.selection_strategy
        self.adaptation_strategy = settings_repository.strategy_settings.adaptation_strategy

        self.ga_settings_dict = vars(settings_repository.ga_settings)
        self.mutation_settings_dict = vars(settings_repository.mutation_settings)
        self.crossover_settings_dict = vars(settings_repository.crossover_settings)
        self.selection_settings_dict = vars(settings_repository.selection_settings)
        self.adaptation_settings_dict = vars(settings_repository.adaptation_settings)

        self.root_individual = population_repository.get_root_individual()

    def execute(self, parents: List[Individual]) -> List[Individual]:
        """Create next generation using parents."""
        average_cost, minimum_cost = self.compute_statistics(parents)

        children = self.initialize(parents)
        while len(children) < self.ga_settings.population_size:
            parent1 = self.select(parents)
            parent2 = self.select(parents)

            self.adapt_cross_and_mutate_rates(average_cost, minimum_cost, (parent1.cost, parent2.cost))

            child1, child2 = self.cross(parent1, parent2)
            child1, child2 = self.mutate(child1, child2)

            children.extend([child1, child2])

        return children

    def initialize(self, parents: List[Individual]) -> List[Individual]:
        children = []
        if self.ga_settings.use_elitism:
            children.extend(sorted(parents)[0:2])
        return children

    def select(self, parents: List[Individual]) -> Individual:
        return self.selection_strategy.selection(parents, **self.selection_settings_dict)

    def cross(self, parent1: Individual, parent2: Individual) -> Tuple[Individual, Individual]:
        child1, child2 = parent1, parent2
        if random.random() < self.crossover_rate:
            child1, child2 = self.crossover_strategy.crossover(parent1, parent2, self.root_individual,
                                                               **self.crossover_settings_dict)
        return child1, child2

    def mutate(self, parent1: Individual, parent2: Individual) -> Tuple[Individual, Individual]:
        child1, child2 = parent1, parent2
        if random.random() < self.mutation_rates[0]:
            child1 = self.mutation_strategy.mutation(parent1, self.root_individual, **self.mutation_settings_dict)
        if random.random() < self.mutation_rates[1]:
            child2 = self.mutation_strategy.mutation(parent2, self.root_individual, **self.mutation_settings_dict)
        return child1, child2

    def adapt_cross_and_mutate_rates(self, average_cost: float, minimum_cost: float, parent_costs: Tuple[float, float]):
        if self.ga_settings.use_adaptation:
            self.crossover_rate, self.mutation_rates = \
                self.adaptation_strategy.adaptation(average_cost, minimum_cost, parent_costs,
                                                    **self.adaptation_settings_dict, **self.ga_settings_dict)

    @staticmethod
    def compute_statistics(population: List[Individual]) -> Tuple[float, float]:
        costs = [individual.cost for individual in population]
        return statistics.mean(costs), min(costs)



#!/usr/bin/env python

"""Factory for error calculation algorithms allowed for usage."""

# Standard library

# 3rd party packages

# Local source
from parametrization_clean.domain.cost.strategy import IErrorStrategy
from parametrization_clean.domain.cost.reax_error import ReaxError


class ErrorFactory:
    """Factory class for creating error calculator algorithm executor - RegistryHolder design pattern.
    Classes that implement IErrorStrategy can be registered and utilized through this factory's registry.
    """

    REGISTRY = {}
    """Internal registry for available error calculation methods. Users can specify from one of the
    `algorithm_name` strings available in the dictionary, mapping `algorithm_name` to the corresponding class
    implementing that algorithm.
    For example, "reax_error" maps to the ReaxFF error calculation algorithm;
    users can specify the `error_strategy` in the user config.json file to use this algorithm.
    """

    @classmethod
    def register(cls, algorithm_name: str, error_calculator_class):
        """Register an error calculation strategy with a string key. Useful for abstraction and dynamic retrieval
        of different algorithms in configuration file. Using this factory, one can easily implement an error
        calculation algorithm (ex. MyErrorCalculatorClass) that follows IErrorStrategy, then use
        "ErrorFactory.register('my_error_calculator_class')"
        to generate a corresponding string reference for that error calculation strategy.

        Parameters
        ----------
        algorithm_name: str
            Name that one wishes to assign to the designated `error_calculator_class`/algorithm.
        error_calculator_class
            Class that one wishes to associate/register with `algorithm_name`.
        Returns
        -------
        error_calculator_class
            Same as the `error_calculator_class` input parameter.
        """
        cls.REGISTRY[algorithm_name] = error_calculator_class
        return error_calculator_class

    @classmethod
    def create_executor(cls, algorithm_name: str) -> IErrorStrategy:
        return cls.REGISTRY[algorithm_name]


ErrorFactory.register('reax_error', ReaxError)



#!/usr/bin/env python

"""Module with class to structure/maintain genetic_algorithm population's "individual"/child."""

# Standard library
from typing import List
from copy import deepcopy

# 3rd party packages

# Local source
from parametrization_clean.domain.root_individual import RootIndividual
from parametrization_clean.domain.cost.strategy import IErrorStrategy
from parametrization_clean.domain.cost.reax_error import ReaxError
from parametrization_clean.domain.utils.helpers import set_param


class Individual(object):

    def __init__(self, params: List[float], reax_energies: List[float] = None,
                 root_individual: RootIndividual = None, error_calculator: IErrorStrategy = ReaxError):
        """Individual/Case in the Genetic Algorithm. Unique based on its own params and corresponding ffield.
        Used as a data storage object.
        Anything requiring creation of updating parameters also requires updating the ffield dict.
        The ffield dict is updated by using the reference/root ffield dict and setting the parameters at the given
        parameter keys.

        Parameters
        ----------
        params: List[float]
            List containing parameter values mapped from param_keys to ffield.
        reax_energies: List[float]
            List containing ReaxFF energies (output from fort.99).
        root_individual: RootIndividual
            Root individual with stored reference/training set configuration: dft energies, weights, etc.
        error_calculator: IErrorStrategy, default = ReaxError
            Strategy for error computation.
        """
        self.params = params
        self.reax_energies = reax_energies

        self.error_calculator = error_calculator
        self.ffield = self.update_ffield(root_individual) if root_individual else None
        self.cost = self.total_error(root_individual) if root_individual and reax_energies else None

    def total_error(self, root_individual: RootIndividual) -> float:
        return sum(self.error_calculator.error(reax_val, dft_val, weight) for reax_val, dft_val, weight in
                   zip(self.reax_energies, root_individual.dft_energies, root_individual.weights))

    def update_ffield(self, root_individual: RootIndividual):
        """Update Individual's `ffield` based on `params` (to be used after mutation/mating).
        Uses deepcopy to avoid mutating the original root ffield.
        """
        ffield = deepcopy(root_individual.root_ffield)
        for param, key in zip(self.params, root_individual.param_keys):
            set_param(key, ffield, param)
        return ffield

    @classmethod
    def from_root_individual(cls, root_individual: RootIndividual):
        """Constructor from RootIndividual.
        Same functionality as calling Individual(params=root_individual.params).
        """
        return cls(params=root_individual.root_params, root_individual=root_individual)

    # For usage of min/max functions in determining best/worst individuals
    def __eq__(self, other):
        return self.cost == other.cost

    def __lt__(self, other):
        return self.cost < other.cost

    def __gt__(self, other):
        return self.cost > other.cost



#!/usr/bin/env python

# Standard library
from typing import Tuple, List
import math

# 3rd party packages


# Local source
from parametrization_clean.domain.adaptation.strategy import IAdaptationStrategy


class XiaoAdapt(IAdaptationStrategy):

    @staticmethod
    def adaptation(average_cost: float, minimum_cost: float, parent_costs: Tuple[float, float], **kwargs) \
            -> Tuple[float, List[float]]:
        """Execute Xiao adaptive algorithm based on the arctan function.

        :param average_cost: Measure of central tendency of cost of population. Can also be the median.
        :param minimum_cost: Lowest cost of population of Individuals.
        :param parent_costs: Tuple containing costs associated with two parents.
        :return: Tuple with (probability of crossover, [probabilities of mutation]).
        """
        max_crossover_probability = kwargs.get('crossover_rate', None)
        max_mutation_probability = kwargs.get('mutation_rate', None)
        min_crossover_probability = kwargs.get('xiao_min_crossover_rate', max_crossover_probability / 2)
        min_mutation_probability = kwargs.get('xiao_min_mutation_rate', max_mutation_probability / 2)
        scale_factor = kwargs.get('xiao_scale', 0.4)

        spread = minimum_cost - average_cost
        better_parent_cost = min(parent_costs)
        if better_parent_cost < average_cost:
            cross_rate = (min_crossover_probability + max_crossover_probability) / 2 \
                         + (min_crossover_probability - max_crossover_probability) / math.pi \
                         * math.atan(scale_factor * (2 * better_parent_cost - spread) / (2 * spread))
        else:
            cross_rate = max_crossover_probability

        mutation_rates = []
        for cost in parent_costs:
            if cost < average_cost:
                mutation_rate = (min_mutation_probability + max_mutation_probability) / 2 \
                                + (min_mutation_probability - max_mutation_probability) / math.pi \
                                * math.atan(scale_factor * (2 * cost - spread) / (2 * spread))
            else:
                mutation_rate = max_mutation_probability
            mutation_rates.append(mutation_rate)

        return cross_rate, mutation_rates




# Standard library
from unittest.mock import patch

# 3rd party packages
import pytest

# Local source
from parametrization_clean.domain.crossover.two_point import TwoPointCross
from tests.domain.crossover.test_double_pareto import get_root_individual


@patch('random.sample')
@pytest.mark.usefixtures('get_individuals')
def test_two_point(sample_mock, get_root_individual, get_individuals):
    sample_mock.return_value = [2, 4]
    parent1 = get_individuals[0]
    parent2 = get_individuals[1]
    child1, child2 = TwoPointCross.crossover(parent1, parent2, root_individual=get_root_individual)
    assert child1.params == [1.0, 0.1, -0.76, 8.6, 2.4]
    assert child2.params == [0.5, 0.05, -0.5, 4.3, 2.1]




# Standard library

# 3rd party packages
import pytest

# Local source
from parametrization_clean.infrastructure.utils.reax_reader import ReaxReader
from parametrization_clean.infrastructure.utils.reax_converter import Fort99Extractor


@pytest.fixture()
@pytest.mark.usefixtures("training_set_dir_path")
def fort99_data(training_set_dir_path):
    reax_reader = ReaxReader(training_set_dir_path)
    return reax_reader.read_fort99()


@pytest.fixture()
def fort99_extractor(fort99_data):
    return Fort99Extractor(fort99_data)


def test_get_reax_energies(fort99_extractor):
    reax_energies = fort99_extractor.get_reax_energies()
    assert isinstance(reax_energies, list)
    assert reax_energies[0] == 1.1183
    assert reax_energies[-1] == -0.7802
    assert reax_energies[180] == 1.4460


def test_get_dft_energies(fort99_extractor):
    dft_energies = fort99_extractor.get_dft_energies()
    assert isinstance(dft_energies, list)
    assert dft_energies[0] == 1.0950
    assert dft_energies[-1] == -0.0010
    assert dft_energies[180] == 0.1090


def test_get_weights(fort99_extractor):
    weights = fort99_extractor.get_weights()
    assert isinstance(weights, list)
    assert weights[0] == 0.0500
    assert weights[-1] == 2.0000
    assert weights[180] == 1.0000



#!/usr/bin/env python

"""Contains default config classes with default parameters used for genetic algorithm/neural network, based on the
settings repository port. Note that the attributes implemented map one-to-one with the corresponding JSON key
that would be expected from the user config, if specified. For example, at the highest level, DefaultSettings
is composed of `strategy_settings`, `ga_settings`, ..., `neural_net_settings`. Settings for each of those attributes
would be specified in the JSON as following:

{
    "strategy_settings": {
        ...
    },
    "ga_settings": {
        ...
    },
    ...
    "neural_net_settings": {
        ...
    }
}

For each of those keys in the JSON, then, the corresponding attributes also map directly to the JSON key names expected.
For example, `ga_settings` represents an implementation of IGeneticAlgorithmSettings, which expects several attributes:
`population_size`, `mutation_rate`, `crossover_rate`, and so on. If the user wishes to change, for example, the
population size, their JSON configuration file might look like

{
    "ga_settings": {
        "population_size": 100
    }
}

The user can choose to override as many of the existing attributes as they wish, but reasonable defaults are provided
for the adaptation, selection, crossover, and mutation algorithms, as well as for the neural network. However, IT IS
HIGHLY RECOMMENDED TO SET, AT THE VERY LEAST, THE POPULATION SIZE OF THE GENETIC ALGORITHM BASED ON THE USER'S DESIRES.
"""

# Standard library

# 3rd party packages

# Local source
from parametrization_clean.domain.selection.tournament import TournamentSelect
from parametrization_clean.domain.mutation.gauss import GaussianMutate
from parametrization_clean.domain.crossover.double_pareto import DoubleParetoCross
from parametrization_clean.domain.adaptation.xiao import XiaoAdapt
from parametrization_clean.domain.cost.reax_error import ReaxError
from parametrization_clean.domain.mutation.nakata import NakataMutate
from parametrization_clean.use_case.port.settings_repository import (IStrategySettings,
                                                                     IGeneticAlgorithmSettings,
                                                                     IMutationSettings,
                                                                     ICrossoverSettings,
                                                                     ISelectionSettings,
                                                                     IAdaptationSettings,
                                                                     INeuralNetSettings,
                                                                     IAllSettings)


class DefaultSettings(IAllSettings):

    def __init__(self):
        super().__init__()
        self.strategy_settings = DefaultStrategySettings()
        self.ga_settings = DefaultGeneticAlgorithmSettings()
        self.mutation_settings = DefaultMutationSettings()
        self.crossover_settings = DefaultCrossoverSettings()
        self.selection_settings = DefaultSelectionSettings()
        self.adaptation_settings = DefaultAdaptationSettings()
        self.neural_net_settings = DefaultNeuralNetSettings()


class DefaultStrategySettings(IStrategySettings):

    def __init__(self):
        super().__init__()
        self.selection_strategy = TournamentSelect
        self.mutation_strategy = GaussianMutate
        self.crossover_strategy = DoubleParetoCross
        self.adaptation_strategy = XiaoAdapt
        self.error_strategy = ReaxError
        self.initialization_strategy = NakataMutate


class DefaultGeneticAlgorithmSettings(IGeneticAlgorithmSettings):

    def __init__(self):
        super().__init__()
        self.population_size = 30
        self.mutation_rate = 0.2
        self.crossover_rate = 0.8
        self.use_elitism = True
        self.use_adaptation = False
        self.use_neural_network = False


class DefaultMutationSettings(IMutationSettings):

    def __init__(self):
        super().__init__()
        self.gauss_std = [0.01, 0.1]
        self.gauss_frac = [0.5, 0.5]

        self.nakata_rand_lower = -1.0
        self.nakata_rand_higher = 1.0
        self.nakata_scale = 0.1

        self.polynomial_eta = 60

        # TODO: Currently causes issues with central uniform mutate -> no mutations caused...
        # Either set equal to None or raise Exception if central uniform mutate called with param bounds that aren't set
        self.param_bounds = []


class DefaultCrossoverSettings(ICrossoverSettings):

    def __init__(self):
        super().__init__()
        self.dpx_alpha = 10
        self.dpx_beta = 1


class DefaultSelectionSettings(ISelectionSettings):

    def __init__(self):
        super().__init__()
        self.tournament_size = 2


class DefaultAdaptationSettings(IAdaptationSettings):

    def __init__(self):
        super().__init__()
        self.srinivas_k1 = 1.0
        self.srinivas_k2 = 0.5
        self.srinivas_k3 = 1.0
        self.srinivas_k4 = 0.5
        self.srinivas_default_mutation_rate = 0.01

        self.xiao_min_mutation_rate = DefaultGeneticAlgorithmSettings().mutation_rate * 0.8
        self.xiao_min_crossover_rate = DefaultGeneticAlgorithmSettings().crossover_rate * 0.8
        self.xiao_scale = 40


class DefaultNeuralNetSettings(INeuralNetSettings):

    def __init__(self):
        super().__init__()
        self.verbosity = 2
        self.train_fraction = 0.8
        self.num_epochs = 10000
        self.num_populations_to_train_on = 1
        self.num_nested_ga_iterations = 1
        self.minimum_validation_r_squared = 0.95



#!/usr/bin/env python

"""Concrete implementation of population repository interface. Gets data through files, since that
is currently the standard method to interact with the standalone ReaxFF optimizations. Also allows output writing
of individuals and entire populations in the genetic algorithm. The `control`, `geo`, `params`, `trainset.in`, and
`ffield` files in the reference training set folder are required. For new individuals generated, an `iopt` file is
automatically generated with one line, a zero, to indicate that ReaxFF should NOT perform its own manual optimization
method, Successive One Parameter Parabolic Extrapolation (SOPPE).

The user has the choice of specifying parameter bounds or leaving them empty in the `params` ReaxFF file.

"""

# Standard library
from typing import List, Tuple
from pathlib import Path
import os
import shutil

# 3rd party packages

# Local source
from parametrization_clean.domain.individual import Individual
from parametrization_clean.domain.root_individual import RootIndividual, FirstGenerationRootIndividual
from parametrization_clean.domain.cost.reax_error import ReaxError
from parametrization_clean.domain.utils.helpers import get_param
from parametrization_clean.use_case.port.population_repository import IPopulationRepository
from parametrization_clean.use_case.port.settings_repository import IAllSettings
from parametrization_clean.infrastructure.utils.reax_reader import ReaxReader, write_ffield
from parametrization_clean.infrastructure.utils.reax_converter import Fort99Extractor
from parametrization_clean.infrastructure.utils.response_object import (ResponseSuccess, ResponseWarning,
                                                                        ResponseFailure)


class PopulationFileRepository(IPopulationRepository):
    GENERATION_FOLDER_PREFIX = "generation-"
    INDIVIDUAL_FOLDER_PREFIX = "child-"
    REFERENCE_FOLDER_PREFIX = "reference-files"

    def __init__(self, training_set_path, population_path, settings_repository: IAllSettings,
                 current_generation_number):
        self.training_set_path = Path(training_set_path)
        self.population_path = Path(population_path)
        self.reference_path = os.path.join(population_path, self.REFERENCE_FOLDER_PREFIX)

        self.training_reax_reader = ReaxReader(training_set_path)
        self.population_reax_reader = ReaxReader(population_path)

        self.population_size = settings_repository.ga_settings.population_size
        self.current_generation_number = current_generation_number

        # SETTING CONFIG's PARAMETER BOUNDS HERE FOR NOW
        self.param_keys, _, param_bounds = self.training_reax_reader.read_params()
        settings_repository.mutation_settings.param_bounds = param_bounds

        if not os.path.isdir(population_path):
            os.mkdir(population_path)

        _, self.atom_types = self.training_reax_reader.read_ffield()

    def get_root_individual(self) -> RootIndividual:
        root_ffield, _ = self.training_reax_reader.read_ffield()

        if self.current_generation_number == 1:
            root_individual = FirstGenerationRootIndividual(root_ffield, self.param_keys)
        else:
            if self.current_generation_number == 2:
                fort99_data, child_dir = self.__find_valid_initial_fort99()
                self.__cache_reference_fort99(child_dir)
            else:
                fort99_data = self.__read_reference_fort99()

            fort99_extractor = Fort99Extractor(fort99_data)
            dft_energies = fort99_extractor.get_dft_energies()
            weights = fort99_extractor.get_weights()

            root_individual = RootIndividual(dft_energies, weights, root_ffield, self.param_keys)

        return root_individual

    def get_population(self, generation_number: int) -> Tuple[List[Individual], List[int]]:
        generation_dir_path = os.path.join(self.population_path, self.GENERATION_FOLDER_PREFIX + str(generation_number))
        root_individual = self.get_root_individual()

        population = []
        successfully_retrieved_case_numbers = []
        for case_number in range(self.population_size):
            child_dir = os.path.join(generation_dir_path, self.INDIVIDUAL_FOLDER_PREFIX + str(case_number))
            self.population_reax_reader.dir_path = child_dir

            try:
                child_fort99_data = self.population_reax_reader.read_fort99()
                child_ffield, _ = self.population_reax_reader.read_ffield()
                child_params = [get_param(key, child_ffield) for key in self.param_keys]

                fort99_extractor = Fort99Extractor(child_fort99_data)
                child_reax_energies = fort99_extractor.get_reax_energies()

                # noinspection PyTypeChecker
                child = Individual(child_params, child_reax_energies, root_individual, error_calculator=ReaxError)

                population.append(child)
                successfully_retrieved_case_numbers.append(case_number)

            except FileNotFoundError:
                # 'fort.99' file does not exist
                print("fort.99 not found in '{}'...continuing to next case".format(child_dir))
                continue
            except ValueError:
                # Invalid number of columns in fort.99 -> skip and continue to next case
                # TODO: Log
                continue

        return population, successfully_retrieved_case_numbers

    def get_previous_n_populations(self, num_populations: int) -> List[Individual]:
        """Read previous N populations before the current generation number.
        Precaution - ensure that lowest possible generation number is zero.
        """
        lower_bound = max(1, self.current_generation_number - num_populations)
        return self.read_population_range(lower_bound=lower_bound, upper_bound=self.current_generation_number)

    def write_individual(self, individual: Individual, **kwargs):
        """Write a `case` in the population to `child_dir`.

        NOTE: If `child_dir` does not exist, it is created.
        NOTE: control, exe, geo, params, trainset.in files are copied from training set directory to `child_dir`.
        NOTE: '0' is written to `iopt` in `child_dir` due to ReaxFF requirement.
        """
        child_dir = kwargs.get('child_dir', None)
        if not child_dir:
            return ResponseFailure.build_parameters_error(message="Child directory not provided.")

        files_overwritten = True
        if not os.path.isdir(child_dir):
            os.mkdir(child_dir)
            files_overwritten = False

        shutil.copy(os.path.join(self.training_set_path, 'control'), child_dir)
        shutil.copy(os.path.join(self.training_set_path, 'geo'), child_dir)
        shutil.copy(os.path.join(self.training_set_path, 'params'), child_dir)
        shutil.copy(os.path.join(self.training_set_path, 'trainset.in'), child_dir)

        with open(os.path.join(child_dir, 'iopt'), 'w') as out_file:
            out_file.write('0')

        try:
            write_ffield(os.path.join(self.training_set_path, 'ffield'), os.path.join(child_dir, 'ffield'),
                         individual.ffield, self.atom_types)
        except (OSError, IOError):
            return ResponseFailure.build_resource_error(
                message="I/O problem with creating child directory at {}".format(child_dir))
        except ValueError:
            return ResponseFailure.build_resource_error(
                message="Trouble writing ffield file in child directory {}".format(child_dir))

        if files_overwritten:
            return ResponseWarning(message="Files in child directory {} were overwritten."
                                   .format(child_dir))
        else:
            return ResponseSuccess(message="Child successfully written at {}.".format(child_dir))

    def write_population(self, population: List[Individual], generation_number):
        """Write a `population` of individuals to the population generation path."""
        generation_dir_path = os.path.join(self.population_path, self.GENERATION_FOLDER_PREFIX + str(generation_number))

        files_overwritten = True
        if not os.path.isdir(generation_dir_path):
            os.mkdir(generation_dir_path)
            files_overwritten = False

        responses = []
        for i, individual in enumerate(population):
            child_dir = os.path.join(generation_dir_path, self.INDIVIDUAL_FOLDER_PREFIX + str(i))
            responses.append(self.write_individual(individual, child_dir=child_dir))

        if any(responses):
            if files_overwritten:
                return ResponseWarning(message="Population generation directory {} overwritten."
                                       .format(generation_dir_path))
            else:
                return ResponseSuccess(message="Generation successfully written at {}"
                                       .format(generation_dir_path))
        else:
            return ResponseFailure.build_resource_error(
                message="No children in population directory {} were written successfully."
                .format(generation_dir_path))

    def read_population_range(self, lower_bound: int, upper_bound: int) -> List[Individual]:
        """Read population of Individuals from `_generation_path` for generations between
        [lower_bound, upper_bound) (lower bound inclusive, upper bound exclusive).
        """
        population = []
        for generation_number in range(lower_bound, upper_bound):
            generation_population, _ = self.get_population(generation_number)
            population.extend(generation_population)
        return population

    def __read_reference_fort99(self) -> List:
        self.training_reax_reader.dir_path = self.reference_path
        fort99_data = self.training_reax_reader.read_fort99()
        self.training_reax_reader.dir_path = self.training_set_path
        return fort99_data

    def __cache_reference_fort99(self, child_dir: str):
        """Takes the fort.99 file found in Generation 1 in the specified child's directory
        and caches it in the population output directory for future usage.
        """
        if not os.path.isdir(self.reference_path):
            os.mkdir(self.reference_path)

        reference_fort99_path = os.path.join(self.reference_path, "fort.99")
        shutil.copy(os.path.join(child_dir, 'fort.99'), reference_fort99_path)

    def __find_valid_initial_fort99(self) -> Tuple[List, str]:
        """Go through Generation 1 and look until a completed child/case with a valid fort.99 file has been found.
        Upon finding the child, retrieve the fort99 data for that child.
        """
        generation_dir_path = os.path.join(self.population_path, self.GENERATION_FOLDER_PREFIX + "1")

        child_dir = ""
        fort99_data = []
        for case_number in range(self.population_size):
            child_dir = os.path.join(generation_dir_path, self.INDIVIDUAL_FOLDER_PREFIX + str(case_number))
            self.training_reax_reader.dir_path = child_dir

            try:
                fort99_data = self.training_reax_reader.read_fort99()
                break
            except FileNotFoundError:
                continue
            except ValueError:
                # Invalid number of columns in fort.99 -> skip and continue to next case
                continue

        return fort99_data, child_dir




# Standard library
from unittest import mock

# 3rd party packages
import pytest

# Local source
from parametrization_clean.use_case.population_writer import PopulationWriter


@mock.patch('parametrization_clean.use_case.port.population_repository.IPopulationRepository')
def test_population_writer_init(repository_mock):
    population_writer = PopulationWriter(population_repository=repository_mock)
    assert population_writer.population_repository == repository_mock


@mock.patch('parametrization_clean.use_case.port.population_repository.IPopulationRepository')
@pytest.mark.usefixtures('get_individuals')
def test_population_writer_write_individual(repository_mock, get_individuals):
    population_writer = PopulationWriter(population_repository=repository_mock)
    population_writer.write_individual(get_individuals[0], child_dir="test/directory/child-1/")
    repository_mock.write_individual.assert_called_once_with(get_individuals[0], child_dir="test/directory/child-1/")


@mock.patch('parametrization_clean.use_case.port.population_repository.IPopulationRepository')
@pytest.mark.usefixtures('get_individuals')
def test_population_writer_write_population(repository_mock, get_individuals):
    population_writer = PopulationWriter(population_repository=repository_mock)
    population_writer.write_population(get_individuals, generation_number=1)
    repository_mock.write_population.assert_called_once_with(get_individuals, 1)



#!/usr/bin/env python

"""Module to parse user-provided, local parameters used for genetic algorithm/neural network.
Overrides default parameters. User-provided configuration file is optional.
"""

# Standard library
import sys
import json

# 3rd party packages

# Local source
from parametrization_clean.infrastructure.config.default import DefaultSettings
from parametrization_clean.infrastructure.exception.exception import ConfigurationError
# Note: the following imports are required so that python can use the relevant algorithm factory in finding the
# user-requested algorithm
from parametrization_clean.domain.selection.factory import SelectionFactory
from parametrization_clean.domain.adaptation.factory import AdaptationFactory
from parametrization_clean.domain.cost.factory import ErrorFactory
from parametrization_clean.domain.crossover.factory import CrossoverFactory
from parametrization_clean.domain.mutation.factory import MutationFactory


# TODO: Need a way to set param bounds, which is required for central uniform mutation
class UserSettings(DefaultSettings):

    def __init__(self, user_config_file_path):
        super().__init__()

        try:
            with open(user_config_file_path, 'r') as in_file:
                all_settings_dict = json.load(in_file)
            self.build_from(all_settings_dict)
        except json.JSONDecodeError:
            raise ConfigurationError("ERROR parsing JSON configuration file at {}."
                                     "Please double check the file and try again.".format(user_config_file_path))
        except FileNotFoundError:
            # No configuration file provided - using default config values
            # TODO: LOG
            pass

    def build_from(self, all_settings_dict):
        self.set_strategy_settings(all_settings_dict.get('strategy_settings', {}))
        self.set_ga_settings(all_settings_dict.get('ga_settings', {}))
        self.set_mutation_settings(all_settings_dict.get('mutation_settings', {}))
        self.set_crossover_settings(all_settings_dict.get('crossover_settings', {}))
        self.set_selection_settings(all_settings_dict.get('selection_settings', {}))
        self.set_adaptation_settings(all_settings_dict.get('adaptation_settings', {}))
        self.set_neural_net_settings(all_settings_dict.get('neural_net_settings', {}))

    def set_strategy_settings(self, strategy_settings_dict):
        for key, value in strategy_settings_dict.items():
            if key == "initialization":  # initialization and mutation share same factory
                factory_name = "MutationFactory"
            else:
                factory_name = key.capitalize() + "Factory"

            attribute_name = key + "_strategy"
            factory = getattr(sys.modules[__name__], factory_name)
            setattr(self.strategy_settings, attribute_name, factory.create_executor(value))

    def set_ga_settings(self, ga_settings_dict):
        self.set_attributes_from_json(self.ga_settings, ga_settings_dict)

    def set_mutation_settings(self, mutation_settings_dict):
        self.set_attributes_from_json(self.mutation_settings, mutation_settings_dict)

    def set_crossover_settings(self, crossover_settings_dict):
        self.set_attributes_from_json(self.crossover_settings, crossover_settings_dict)

    def set_selection_settings(self, selection_settings_dict):
        self.set_attributes_from_json(self.selection_settings, selection_settings_dict)

    def set_adaptation_settings(self, adaptation_settings_dict):
        self.set_attributes_from_json(self.adaptation_settings, adaptation_settings_dict)

    def set_neural_net_settings(self, neural_net_settings_dict):
        self.set_attributes_from_json(self.neural_net_settings, neural_net_settings_dict)

    @staticmethod
    def set_attributes_from_json(config_object, json_dict):
        """Extract and set class attributes from JSON dictionary object."""
        for key, value in json_dict.items():
            attribute_name = key
            setattr(config_object, attribute_name, value)




# Standard library
from unittest import mock

# 3rd party packages
import pytest

# Local source
from parametrization_clean.domain.root_individual import RootIndividual
from parametrization_clean.domain.mutation.nakata import NakataMutate
from parametrization_clean.domain.mutation.central_uniform import CentralUniformMutate
from parametrization_clean.use_case.population_initializer import PopulationInitializer
from tests.use_case.test_population_propagator import all_settings


@pytest.fixture()
@pytest.mark.usefixtures('dft_energies', 'weights', 'root_ffield', 'param_keys')
def root_individual(dft_energies, weights, root_ffield, param_keys):
    root_individual = RootIndividual(dft_energies, weights, root_ffield, param_keys)
    return root_individual


@mock.patch('parametrization_clean.use_case.port.population_repository.IPopulationRepository')
def test_population_initializer_init(population_repository_mock, all_settings, root_individual):
    all_settings.initialization_strategy = NakataMutate
    population_repository_mock.get_root_individual = mock.MagicMock(return_value=root_individual)
    initializer = PopulationInitializer(population_repository=population_repository_mock,
                                        settings_repository=all_settings)
    assert initializer.population_repository == population_repository_mock
    assert initializer.population_size == 4

    assert initializer.mutation_settings_dict['gauss_std'] == [0.10]
    assert initializer.mutation_settings_dict['gauss_frac'] == [1.0]
    assert initializer.mutation_settings_dict['nakata_rand_lower'] == -1.0
    assert initializer.mutation_settings_dict['nakata_rand_higher'] == 1.0
    assert initializer.mutation_settings_dict['nakata_scale'] == 0.10
    assert initializer.mutation_settings_dict['polynomial_eta'] == 60
    assert initializer.mutation_settings_dict['param_bounds'] == [[0.0, 2.0],
                                                                  [0.0, 0.5],
                                                                  [-1.0, 0.0],
                                                                  [3.0, 12.0],
                                                                  [1.0, 3.0]]


@mock.patch('parametrization_clean.use_case.port.population_repository.IPopulationRepository')
def test_population_initializer(population_repository_mock, all_settings, root_individual):
    all_settings.initialization_strategy = NakataMutate
    population_repository_mock.get_root_individual = mock.MagicMock(return_value=root_individual)
    initializer = PopulationInitializer(population_repository=population_repository_mock,
                                        settings_repository=all_settings)
    population = initializer.execute()
    assert len(population) == 4
    for case in population:
        assert case.params != root_individual.root_params


@mock.patch('parametrization_clean.use_case.port.population_repository.IPopulationRepository')
@pytest.mark.usefixtures('param_bounds')
def test_population_initializer_with_kwargs(population_repository_mock, all_settings, param_bounds, root_individual):
    all_settings.initialization_strategy = CentralUniformMutate
    all_settings.mutation_settings.param_bounds = param_bounds

    population_repository_mock.get_root_individual = mock.MagicMock(return_value=root_individual)
    initializer = PopulationInitializer(population_repository=population_repository_mock,
                                        settings_repository=all_settings)
    population = initializer.execute()
    assert len(population) == 4
    for case in population:
        assert case.params != root_individual.root_params
        for param, param_bound in zip(case.params, param_bounds):
            assert param_bound[0] <= param <= param_bound[1]



#!/usr/bin/env python

"""Factory for adaptation algorithms allowed for usage."""

# Standard library

# 3rd party packages

# Local source
from parametrization_clean.domain.adaptation.strategy import IAdaptationStrategy
from parametrization_clean.domain.adaptation.srinivas import SrinivasAdapt
from parametrization_clean.domain.adaptation.xiao import XiaoAdapt


class AdaptationFactory:
    """Factory class for creating adaptation algorithm executor - RegistryHolder design pattern.
    Classes that implement IAdaptationStrategy can be registered and utilized through this factory's registry.
    """

    REGISTRY = {}
    """Internal registry for available adaptation methods. Users can specify from one of the `algorithm_name` strings
    available in the dictionary, mapping `algorithm_name` to the corresponding class implementing that algorithm.
    For example, "xiao" maps to Xiao's adaptation algorithm; users can specify the `adaptation_strategy` in the
    user config.json file to use this algorithm.
    """

    @classmethod
    def register(cls, algorithm_name: str, adaptation_class):
        """Register an adaptation strategy with a string key. Useful for abstraction and dynamic retrieval
        of different algorithms in configuration file. Using this factory, one can easily implement an adaptation
        algorithm (ex. MyAdaptationClass) that follows IAdaptationStrategy, then use
        "AdaptationFactory.register('my_adaptation_class')"
        to generate a corresponding string reference for that adaptation strategy.

        Parameters
        ----------
        algorithm_name: str
            Name that one wishes to assign to the designated `adaptation_class`/algorithm.
        adaptation_class
            Class that one wishes to associate/register with `algorithm_name`.
        Returns
        -------
        adaptation_class
            Same as the `adaptation_class` input parameter.
        """
        cls.REGISTRY[algorithm_name] = adaptation_class
        return adaptation_class

    @classmethod
    def create_executor(cls, algorithm_name: str) -> IAdaptationStrategy:
        return cls.REGISTRY[algorithm_name]


AdaptationFactory.register('srinivas', SrinivasAdapt)
AdaptationFactory.register('xiao', XiaoAdapt)




# Standard library

# 3rd party packages

# Local source
from parametrization_clean.domain.mutation.factory import MutationFactory
from parametrization_clean.domain.mutation.nakata import NakataMutate
from parametrization_clean.domain.mutation.central_uniform import CentralUniformMutate
from parametrization_clean.domain.mutation.polynomial import PolynomialMutate
from parametrization_clean.domain.mutation.gauss import GaussianMutate


def test_get_nakata():
    assert MutationFactory.create_executor('nakata') == NakataMutate


def test_get_central_uniform():
    assert MutationFactory.create_executor('central_uniform') == CentralUniformMutate


def test_get_polynomial():
    assert MutationFactory.create_executor('polynomial') == PolynomialMutate


def test_get_gauss():
    assert MutationFactory.create_executor('gauss') == GaussianMutate




# Standard library

# 3rd party packages

# Local source
from parametrization_clean.domain.crossover.factory import CrossoverFactory
from parametrization_clean.domain.crossover.single_point import SinglePointCross
from parametrization_clean.domain.crossover.two_point import TwoPointCross
from parametrization_clean.domain.crossover.uniform import UniformCross
from parametrization_clean.domain.crossover.double_pareto import DoubleParetoCross


def test_get_single_point():
    single_point_cross = CrossoverFactory.create_executor('single_point')
    assert single_point_cross == SinglePointCross


def test_get_two_point():
    assert CrossoverFactory.create_executor('two_point') == TwoPointCross


def test_get_uniform():
    assert CrossoverFactory.create_executor('uniform') == UniformCross


def test_get_double_pareto():
    assert CrossoverFactory.create_executor('double_pareto') == DoubleParetoCross



# Standard library

# 3rd party packages
import pytest
import numpy as np
import pandas as pd

# Local source
from parametrization_clean.domain.neural_network.extract_data import (individuals_to_features_df,
                                                                      individuals_to_features_and_outputs_df)
from parametrization_clean.domain.neural_network.transform_data import (train_test_split, normalize_features,
                                                                        denormalize_features,
                                                                        get_columns_to_remove,
                                                                        remove_problematic_columns)


@pytest.fixture()
@pytest.mark.usefixtures('get_individuals')
def x_df(get_individuals):
    return individuals_to_features_df(get_individuals)


@pytest.fixture()
@pytest.mark.usefixtures('get_individuals')
def x_y_df(get_individuals):
    return individuals_to_features_and_outputs_df(get_individuals)


@pytest.fixture()
def x_train_stats(x_df):
    return x_df.describe().transpose()


def test_train_test_split(x_y_df):
    train_fraction = 0.75
    train_df, test_df = train_test_split(x_y_df, train_fraction)
    assert train_df.shape[0] == 3
    assert test_df.shape[0] == 1
    assert train_df.shape[1] == 9
    assert test_df.shape[1] == 9


def test_normalize_features(x_df, x_train_stats):
    normalized_x_df = normalize_features(x_df, x_train_stats)
    normalized_x_nd_array = normalized_x_df.to_numpy()
    expected_normalized_array = np.array([[0.412005595, -0.488850326, -0.349418121, -1.178611022, 0.968101956],
                                          [-1.142732499, -1.063968358, -1.214643944, 0.483037304, -0.033382826],
                                          [1.15827988, 0.316314917, 1.081532279, -0.405751336, 0.433976739],
                                          [-0.427552976, 1.236503767, 0.482529786, 1.101325054, -1.368695868]
                                          ])
    for param, expected_param in np.nditer([normalized_x_nd_array, expected_normalized_array]):
        assert param == pytest.approx(expected_param)


def test_denormalize_features(x_df, x_train_stats):
    normalized_x_df = normalize_features(x_df, x_train_stats)
    denormalized_x_df = denormalize_features(normalized_x_df, x_train_stats)
    denormalized_x_nd_array = denormalized_x_df.to_numpy()
    expected_denormalized_array = np.array([[1.0, 0.1, -0.5, 4.3, 2.4],
                                            [0.5, 0.05, -0.76, 8.6, 2.1],
                                            [1.24, 0.17, -0.07, 6.3, 2.24],
                                            [0.73, 0.25, -0.25, 10.2, 1.7]
                                            ])
    for param, expected_param in np.nditer([denormalized_x_nd_array, expected_denormalized_array]):
        assert param == pytest.approx(expected_param)


def test_get_columns_to_remove():
    x_df = pd.DataFrame([[1.0, 0.1, -0.5, 4.3, 2.4],
                         [0.5, 0.1, -0.76, 8.6, 2.1],
                         [1.24, 0.1, -0.07, 6.3, 2.24],
                         [0.73, 0.1, -0.25, 10.2, 1.7]
                         ])
    cols_to_remove = get_columns_to_remove(x_df)
    assert cols_to_remove == [1]


def test_remove_problematic_columns():
    x_df = pd.DataFrame([[1.0, 0.1, -0.5, 4.3, 2.4],
                         [0.5, 0.1, -0.76, 8.6, 2.1],
                         [1.24, 0.1, -0.07, 6.3, 2.24],
                         [0.73, 0.1, -0.25, 10.2, 1.7]
                         ])
    columns_to_remove = get_columns_to_remove(x_df)
    new_x_df = remove_problematic_columns(x_df, columns_to_remove)
    new_x_nd_array = new_x_df.to_numpy()
    expected_new_array = np.array([[1.0, -0.5, 4.3, 2.4],
                                   [0.5, -0.76, 8.6, 2.1],
                                   [1.24, -0.07, 6.3, 2.24],
                                   [0.73, -0.25, 10.2, 1.7]
                                   ])
    assert new_x_nd_array.shape == expected_new_array.shape
    for param, expected_param in np.nditer([new_x_nd_array, expected_new_array]):
        assert param == expected_param




# Standard library
from typing import List
from unittest import mock

# 3rd party packages
import pytest

# Local source
from parametrization_clean.domain.individual import Individual
from parametrization_clean.domain.selection.tournament import TournamentSelect
from parametrization_clean.domain.mutation.gauss import GaussianMutate
from parametrization_clean.domain.crossover.double_pareto import DoubleParetoCross
from parametrization_clean.domain.adaptation.xiao import XiaoAdapt
from parametrization_clean.use_case.population_propagator import PopulationPropagator
from parametrization_clean.domain.root_individual import RootIndividual
from parametrization_clean.use_case.port.population_repository import IPopulationRepository
from tests.fixtures.domain import dft_energies, weights, root_ffield, param_keys


@pytest.fixture()
@pytest.mark.usefixtures('dft_energies', 'weights', 'root_ffield', 'param_keys')
def root_individual(dft_energies, weights, root_ffield, param_keys):
    root_individual = RootIndividual(dft_energies, weights, root_ffield, param_keys)
    return root_individual


@pytest.fixture()
def population_repository(root_individual):
    class TestPopulationRepository(IPopulationRepository):
        def get_root_individual(self) -> RootIndividual:
            return root_individual

        def get_population(self, generation_number: int) -> List[Individual]:
            pass

        def get_previous_n_populations(self, num_populations: int) -> List[Individual]:
            pass

        def write_individual(self, individual: Individual, **kwargs):
            pass

        def write_population(self, population: List[Individual], generation_number):
            pass

    return TestPopulationRepository()


@pytest.fixture()
@mock.patch('parametrization_clean.use_case.port.settings_repository.IAllSettings')
@pytest.mark.usefixtures('param_bounds')
def all_settings(all_settings_mock, param_bounds):
    all_settings_mock.strategy_settings.selection_strategy = TournamentSelect
    all_settings_mock.strategy_settings.mutation_strategy = GaussianMutate
    all_settings_mock.strategy_settings.crossover_strategy = DoubleParetoCross
    all_settings_mock.strategy_settings.adaptation_strategy = XiaoAdapt

    all_settings_mock.ga_settings.population_size = 4
    all_settings_mock.ga_settings.mutation_rate = 0.20
    all_settings_mock.ga_settings.crossover_rate = 0.80
    all_settings_mock.ga_settings.use_elitism = True
    all_settings_mock.ga_settings.use_adaptation = False
    all_settings_mock.ga_settings.use_neural_network = False

    all_settings_mock.mutation_settings.gauss_std = [0.10]
    all_settings_mock.mutation_settings.gauss_frac = [1.0]
    all_settings_mock.mutation_settings.nakata_rand_lower = -1.0
    all_settings_mock.mutation_settings.nakata_rand_higher = 1.0
    all_settings_mock.mutation_settings.nakata_scale = 0.10
    all_settings_mock.mutation_settings.polynomial_eta = 60
    all_settings_mock.mutation_settings.param_bounds = param_bounds

    all_settings_mock.crossover_settings.dpx_alpha = 10
    all_settings_mock.crossover_settings.dpx_beta = 1

    all_settings_mock.selection_settings.tournament_size = 2

    all_settings_mock.adaptation_settings.srinivas_k1 = 1.0
    all_settings_mock.adaptation_settings.srinivas_k2 = 0.5
    all_settings_mock.adaptation_settings.srinivas_k3 = 1.0
    all_settings_mock.adaptation_settings.srinivas_k4 = 0.5
    all_settings_mock.adaptation_settings.srinivas_default_mutation_rate = 0.005
    all_settings_mock.adaptation_settings.xiao_min_crossover_rate = 0.4
    all_settings_mock.adaptation_settings.xiao_min_mutation_rate = 0.1
    all_settings_mock.adaptation_settings.xiao_scale = 0.4

    return all_settings_mock


@mock.patch('parametrization_clean.use_case.port.population_repository.IPopulationRepository')
def test_population_propagator_init(population_repository_mock, all_settings):
    population_repository_mock.get_root_individual = mock.MagicMock(return_value=root_individual)

    propagator = PopulationPropagator(all_settings, population_repository_mock)

    assert propagator.ga_settings.population_size == 4
    assert propagator.ga_settings.mutation_rate == 0.20
    assert propagator.ga_settings.crossover_rate == 0.80
    assert propagator.ga_settings.use_elitism
    assert not propagator.ga_settings.use_adaptation
    assert not propagator.ga_settings.use_neural_network

    assert propagator.crossover_rate == 0.80
    assert propagator.mutation_rates == [0.20, 0.20]

    assert propagator.mutation_strategy == all_settings.strategy_settings.mutation_strategy
    assert propagator.crossover_strategy == all_settings.strategy_settings.crossover_strategy
    assert propagator.selection_strategy == all_settings.strategy_settings.selection_strategy

    assert propagator.mutation_settings_dict['gauss_std'] == [0.10]
    assert propagator.mutation_settings_dict['gauss_frac'] == [1.0]
    assert propagator.mutation_settings_dict['nakata_rand_lower'] == -1.0
    assert propagator.mutation_settings_dict['nakata_rand_higher'] == 1.0
    assert propagator.mutation_settings_dict['nakata_scale'] == 0.10
    assert propagator.mutation_settings_dict['polynomial_eta'] == 60
    assert propagator.mutation_settings_dict['param_bounds'] == [[0.0, 2.0],
                                                                 [0.0, 0.5],
                                                                 [-1.0, 0.0],
                                                                 [3.0, 12.0],
                                                                 [1.0, 3.0]]

    assert propagator.crossover_settings_dict['dpx_alpha'] == 10
    assert propagator.crossover_settings_dict['dpx_beta'] == 1

    assert propagator.selection_settings_dict['tournament_size'] == 2

    assert propagator.adaptation_settings_dict['srinivas_k1'] == 1.0
    assert propagator.adaptation_settings_dict['srinivas_k2'] == 0.5
    assert propagator.adaptation_settings_dict['srinivas_k3'] == 1.0
    assert propagator.adaptation_settings_dict['srinivas_k4'] == 0.5
    assert propagator.adaptation_settings_dict['xiao_min_crossover_rate'] == 0.4
    assert propagator.adaptation_settings_dict['xiao_min_mutation_rate'] == 0.1
    assert propagator.adaptation_settings_dict['xiao_scale'] == 0.4


@mock.patch('parametrization_clean.use_case.port.population_repository.IPopulationRepository')
@pytest.mark.usefixtures('get_individuals')
def test_population_propagator_initialize(population_repository_mock, get_individuals, root_individual, all_settings):
    population_repository_mock.get_root_individual = mock.MagicMock(return_value=root_individual)

    costs = []
    for individual in get_individuals:
        individual.cost = individual.total_error(root_individual)
        costs.append(individual.cost)
    propagator = PopulationPropagator(all_settings, population_repository_mock)
    sorted_costs = sorted(costs)

    children = propagator.initialize(get_individuals)
    assert len(children) == 2
    assert children[0].cost == sorted_costs[0]
    assert children[1].cost == sorted_costs[1]

    propagator.ga_settings.use_elitism = False
    children = propagator.initialize(get_individuals)
    assert len(children) == 0


@mock.patch('parametrization_clean.use_case.port.population_repository.IPopulationRepository')
@pytest.mark.usefixtures('get_individuals')
def test_population_propagator_select(population_repository_mock, get_individuals, root_individual, all_settings):
    population_repository_mock.get_root_individual = mock.MagicMock(return_value=root_individual)

    for individual in get_individuals:
        individual.cost = individual.total_error(root_individual)
    propagator = PopulationPropagator(all_settings, population_repository_mock)

    selected_individual = propagator.select(get_individuals)
    assert selected_individual in get_individuals


@mock.patch('random.random')
@pytest.mark.usefixtures('get_individuals')
def test_population_propagator_cross(rand_mock, population_repository, get_individuals, all_settings):
    propagator = PopulationPropagator(all_settings, population_repository)
    parent1 = get_individuals[0]
    parent2 = get_individuals[1]

    rand_mock.return_value = 0.60
    child1, child2 = propagator.cross(parent1, parent2)
    assert child1.params != parent1.params and child1.params != parent2.params
    assert child2.params != parent2.params and child2.params != parent1.params

    rand_mock.return_value = 0.90
    child1, child2 = propagator.cross(parent1, parent2)
    assert child1.params == parent1.params and child2.params == parent2.params


@mock.patch('random.random')
@pytest.mark.usefixtures('get_individuals')
def test_population_propagator_mutate(rand_mock, population_repository, get_individuals, all_settings):
    propagator = PopulationPropagator(all_settings, population_repository)
    parent1 = get_individuals[0]
    parent2 = get_individuals[1]

    propagator.mutation_rates = [0.20, 0.20]
    rand_mock.return_value = 0.10
    child1, child2 = propagator.mutate(parent1, parent2)
    assert child1.params != parent1.params
    assert child2.params != parent2.params

    propagator.mutation_rates = [0.20, 0.05]
    child1, child2 = propagator.mutate(parent1, parent2)
    assert child1.params != parent1.params
    assert child2.params == parent2.params


@mock.patch('parametrization_clean.use_case.port.population_repository.IPopulationRepository')
@pytest.mark.usefixtures('get_individuals')
def test_population_propagator_adapt(population_repository_mock, get_individuals, root_individual, all_settings):
    population_repository_mock.get_root_individual = mock.MagicMock(return_value=root_individual)

    for individual in get_individuals:
        individual.cost = individual.total_error(root_individual)

    propagator = PopulationPropagator(all_settings, population_repository_mock)
    propagator.ga_settings.use_adaptation = True

    # Two mock dictionaries passed in kwargs for adaptation...must remove duplicate keys
    propagator.ga_settings_dict.pop('_mock_return_value')
    propagator.ga_settings_dict.pop('_mock_parent')
    propagator.ga_settings_dict.pop('_mock_name')
    propagator.ga_settings_dict.pop('_mock_new_name')
    propagator.ga_settings_dict.pop('_mock_new_parent')
    propagator.ga_settings_dict.pop('_mock_sealed')
    propagator.ga_settings_dict.pop('_spec_class')
    propagator.ga_settings_dict.pop('_spec_set')
    propagator.ga_settings_dict.pop('_spec_signature')
    propagator.ga_settings_dict.pop('_mock_methods')
    propagator.ga_settings_dict.pop('_mock_children')
    propagator.ga_settings_dict.pop('_mock_wraps')
    propagator.ga_settings_dict.pop('_mock_delegate')
    propagator.ga_settings_dict.pop('_mock_called')
    propagator.ga_settings_dict.pop('_mock_call_args')
    propagator.ga_settings_dict.pop('_mock_call_count')
    propagator.ga_settings_dict.pop('_mock_call_args_list')
    propagator.ga_settings_dict.pop('_mock_mock_calls')
    propagator.ga_settings_dict.pop('method_calls')
    propagator.ga_settings_dict.pop('_mock_unsafe')
    propagator.ga_settings_dict.pop('_mock_side_effect')
    # Two mock dictionaries passed in kwargs for adaptation...must remove duplicate keys

    average_cost, minimum_cost = propagator.compute_statistics(get_individuals)
    parent1_cost = get_individuals[2].cost
    parent2_cost = get_individuals[3].cost
    propagator.adapt_cross_and_mutate_rates(average_cost, minimum_cost, (parent1_cost, parent2_cost))
    assert propagator.crossover_rate != 0.80
    assert propagator.mutation_rates[0] != 0.20
    assert propagator.mutation_rates[1] == 0.20


@mock.patch('parametrization_clean.use_case.port.population_repository.IPopulationRepository')
@pytest.mark.usefixtures('get_individuals')
def test_population_propagator_compute_statistics(population_repository_mock, get_individuals,
                                                  root_individual, all_settings):
    population_repository_mock.get_root_individual = mock.MagicMock(return_value=root_individual)

    for individual in get_individuals:
        individual.cost = individual.total_error(root_individual)
    propagator = PopulationPropagator(all_settings, population_repository_mock)
    parents = get_individuals

    average_cost, minimum_cost = propagator.compute_statistics(parents)
    assert average_cost == pytest.approx(1037.955, rel=1e-3)
    assert minimum_cost == pytest.approx(239.558, rel=1e-3)


@mock.patch('parametrization_clean.use_case.port.population_repository.IPopulationRepository')
@pytest.mark.usefixtures('get_individuals')
def test_population_propagator_execute(population_repository_mock, get_individuals, root_individual, all_settings):
    population_repository_mock.get_root_individual = mock.MagicMock(return_value=root_individual)

    for individual in get_individuals:
        individual.cost = individual.total_error(root_individual)
    propagator = PopulationPropagator(all_settings, population_repository_mock)
    parents = get_individuals

    children = propagator.execute(parents)

    assert len(children) == 4



#!/usr/bin/env python

"""Factory for mutation algorithms allowed for usage."""

# Standard library

# 3rd party packages

# Local source
from parametrization_clean.domain.mutation.strategy import IMutationStrategy
from parametrization_clean.domain.mutation.central_uniform import CentralUniformMutate
from parametrization_clean.domain.mutation.nakata import NakataMutate
from parametrization_clean.domain.mutation.gauss import GaussianMutate
from parametrization_clean.domain.mutation.polynomial import PolynomialMutate


class MutationFactory:
    """Factory class for creating mutation algorithm executor - RegistryHolder design pattern.
    Classes that implement IMutationStrategy can be registered and utilized through this factory's registry.
    """

    REGISTRY = {}
    """Internal registry for available mutation methods. Users can specify from one of the
    `algorithm_name` strings available in the dictionary, mapping `algorithm_name` to the corresponding class
    implementing that algorithm.
    For example, "gauss" maps to the multi-factor Gaussian algorithm;
    users can specify the `mutation_strategy` in the user config.json file to use this algorithm.
    """

    @classmethod
    def register(cls, algorithm_name: str, mutation_class):
        """Register a mutation strategy with a string key. Useful for abstraction and dynamic retrieval
        of different algorithms in configuration file. Using this factory, one can easily implement a crossover
        algorithm (ex. MyMutationClass) that follows IMutationStrategy, then use
        "MutationFactory.register('my_mutation_class')"
        to generate a corresponding string reference for that mutation strategy.

        Parameters
        ----------
        algorithm_name: str
            Name that one wishes to assign to the designated `mutation_class`/algorithm.
        mutation_class
            Class that one wishes to associate/register with `algorithm_name`.
        Returns
        -------
        mutation_class
            Same as the `mutation_class` input parameter.
        """
        cls.REGISTRY[algorithm_name] = mutation_class
        return mutation_class

    @classmethod
    def create_executor(cls, algorithm_name: str) -> IMutationStrategy:
        return cls.REGISTRY[algorithm_name]


# Register all existing mutation algorithms
MutationFactory.register('central_uniform', CentralUniformMutate)
MutationFactory.register('nakata', NakataMutate)
MutationFactory.register('gauss', GaussianMutate)
MutationFactory.register('polynomial', PolynomialMutate)



#!/usr/bin/env python

"""Module with class to initialize first generation ReaxFF genetic algorithm by applying mutations to
a reference training case, then output the corresponding first generation of individuals using the population
repository.

PROBLEM: Reference training set must also contain 'fort.99', which is an output file from ReaxFF optimization.
The first generation really shouldn't require the user to already have a fort.99 file; this should be revised in the
future.
"""

# Standard library
from typing import List

# 3rd party packages

# Local source
from parametrization_clean.domain.individual import Individual
from parametrization_clean.use_case.port.population_repository import IPopulationRepository
from parametrization_clean.use_case.port.settings_repository import IAllSettings


class PopulationInitializer(object):

    def __init__(self, population_repository: IPopulationRepository, settings_repository: IAllSettings):
        self.population_repository = population_repository
        self.strategy = settings_repository.strategy_settings.initialization_strategy
        self.population_size = settings_repository.ga_settings.population_size
        self.mutation_settings_dict = vars(settings_repository.mutation_settings)

    def execute(self) -> List[Individual]:
        # TODO: Population initialization shouldn't require dft energies, weights (ex. info from fort.99)
        root_individual = self.population_repository.get_root_individual()
        individual = Individual.from_root_individual(root_individual)
        population = [self.strategy.mutation(individual, root_individual, **self.mutation_settings_dict)
                      for _ in range(self.population_size)]
        return population



#!/usr/bin/env python

"""Module to convert data outputs from ReaxFF I/O object to desired data structures for domain objects."""

# Standard library

# 3rd party packages

# Local source


class Fort99Extractor:

    def __init__(self, fort99_data):
        self.data = fort99_data

    def get_reax_energies(self):
        return [row[0] for row in self.data]

    def get_dft_energies(self):
        return [row[1] for row in self.data]

    def get_weights(self):
        return [row[2] for row in self.data]




# Standard library

# 3rd party packages
import pytest

# Local source
from parametrization_clean.domain.mutation.central_uniform import CentralUniformMutate
from tests.domain.crossover.test_double_pareto import get_root_individual


@pytest.mark.usefixtures('get_individuals', 'param_bounds')
def test_central_uniform(get_root_individual, get_individuals, param_bounds):
    parent = get_individuals[0]
    child = CentralUniformMutate.mutation(parent, root_individual=get_root_individual, param_bounds=param_bounds)
    for param, param_bounds in zip(child.params, param_bounds):
        delta = (param_bounds[1] - param_bounds[0]) / 4
        low = param_bounds[0] + delta
        high = param_bounds[1] - delta
        assert low <= param <= high




# Standard library

# 3rd party packages
import pytest

# Local source
from parametrization_clean.domain.root_individual import RootIndividual


@pytest.mark.usefixtures('dft_energies', 'weights', 'root_ffield', 'param_keys')
def test_root_individual_init(dft_energies, weights, root_ffield, param_keys):
    root_individual = RootIndividual(dft_energies, weights, root_ffield, param_keys)
    true_params = [1, 7, 1, 3, 8]
    assert list(root_individual.dft_energies) == dft_energies
    assert list(root_individual.weights) == weights
    assert root_individual.root_ffield == root_ffield
    assert root_individual.param_keys == param_keys
    assert root_individual.extract_params() == true_params
    assert root_individual.root_params == true_params




# Standard library

# 3rd party packages
import pytest

# Local source
from parametrization_clean.domain.mutation.nakata import NakataMutate
from tests.domain.crossover.test_double_pareto import get_root_individual


@pytest.mark.usefixtures('get_individuals')
def test_single_point(get_root_individual, get_individuals):
    parent = get_individuals[0]
    child = NakataMutate.mutation(parent, get_root_individual,
                                  nakata_scale=0.1, nakata_rand_lower=-1.0, nakata_rand_higher=1.0)
    for param in child.params:
        low = min(0.9 * param, 1.1 * param)
        high = max(0.9 * param, 1.1 * param)
        assert low <= param <= high



# Standard library
import os

# 3rd party packages
import pytest

# Local source
from parametrization_clean.infrastructure.presenter.file_writer import DataWriter
from tests.infrastructure.repository.test_from_files import file_repository
from tests.use_case.test_population_propagator import all_settings
from parametrization_clean.infrastructure.config.default import DefaultSettings
from parametrization_clean.use_case.nested_ga_with_ann import GeneticNeuralNetPropagator


def read_csv(file_path, delimiter=','):
    data = []
    with open(file_path, 'r') as in_file:
        for line in in_file:
            data.append([float(val) for val in line.split(delimiter)])
    return data


@pytest.mark.usefixtures("presenter_output_path")
def test_write_rows_to_csv(file_repository, presenter_output_path):
    csv_file = os.path.join(presenter_output_path, "test_output.csv")

    if os.path.isfile(csv_file):
        os.remove(csv_file)

    first_generation_population, successfully_retrieved_case_numbers = file_repository.get_population(
        generation_number=1)
    costs = [individual.cost for individual in first_generation_population]
    combined_data = [[first_val, second_val] for first_val, second_val in zip(successfully_retrieved_case_numbers,
                                                                              costs)]

    DataWriter.write_rows_to_csv(csv_file, combined_data, delimiter=',')
    read_data = read_csv(csv_file)
    for read_row, expected_row in zip(read_data, combined_data):
        for read_val, expected_val in zip(read_row, expected_row):
            assert read_val == expected_val

    os.remove(csv_file)
    DataWriter.write_rows_to_csv(csv_file, combined_data)
    read_data = read_csv(csv_file)
    for read_row, expected_row in zip(read_data, combined_data):
        for read_val, expected_val in zip(read_row, expected_row):
            assert read_val == expected_val

    os.remove(csv_file)
    DataWriter.write_rows_to_csv(csv_file, combined_data, "\t")
    read_data = read_csv(csv_file, delimiter='\t')
    for read_row, expected_row in zip(read_data, combined_data):
        for read_val, expected_val in zip(read_row, expected_row):
            assert read_val == expected_val


@pytest.mark.usefixtures("presenter_output_path")
def test_create_or_append(file_repository, presenter_output_path):
    file = os.path.join(presenter_output_path, "test_file.txt")

    if os.path.isfile(file):
        os.remove(file)

    first_generation_population, successfully_retrieved_case_numbers = file_repository.get_population(
        generation_number=1)
    costs = [individual.cost for individual in first_generation_population]
    combined_data = [[first_val, second_val] for first_val, second_val in zip(successfully_retrieved_case_numbers,
                                                                              costs)]

    formatted_first_row = "{},{:.3f}\n".format(combined_data[0][0], combined_data[0][1])
    DataWriter.create_or_append(formatted_first_row, file)
    read_data = read_csv(file)
    assert read_data[0][0] == combined_data[0][0]
    assert read_data[0][1] == round(combined_data[0][1], 3)

    formatted_second_row = "{},{:.3f}\n".format(combined_data[1][0], combined_data[1][1])
    DataWriter.create_or_append(formatted_second_row, file)
    read_data = read_csv(file)
    assert read_data[0][0] == combined_data[0][0]
    assert read_data[0][1] == round(combined_data[0][1], 3)
    assert read_data[1][0] == combined_data[1][0]
    assert read_data[1][1] == round(combined_data[1][1], 3)


@pytest.mark.usefixtures("presenter_output_path")
def test_write_summary(file_repository, presenter_output_path):
    summary_file = os.path.join(presenter_output_path, "00-gen-summary.txt")

    if os.path.isfile(summary_file):
        os.remove(summary_file)

    first_generation_population, successfully_retrieved_case_numbers = file_repository.get_population(
        generation_number=1)
    expected_population_size = 5
    DataWriter.write_summary(first_generation_population, expected_population_size,
                             successfully_retrieved_case_numbers, summary_file)

    with open(summary_file, "r") as in_file:
        for _ in range(3):
            next(in_file)
        best_error = float(next(in_file).split()[-1])
        assert best_error == 4094932.470
        best_case = int(next(in_file).split('-')[-1])
        assert best_case == 1


@pytest.mark.usefixtures("presenter_output_path", "reax_output_dir_path")
def test_write_outputs(file_repository, reax_output_dir_path, presenter_output_path):
    previous_generation_number = 1
    file_repository.current_generation_number = 2
    default_settings = DefaultSettings()
    default_settings.neural_net_settings.num_epochs = 1

    summary_file = os.path.join(presenter_output_path, "generation-" + str(previous_generation_number), "00-gen-summary.txt")
    generation_vs_error_file_path = os.path.join(presenter_output_path, '00-generation-vs-error.txt')
    neural_network_file_path = os.path.join(presenter_output_path, "00-ann-summary")

    if os.path.isfile(summary_file):
        os.remove(summary_file)
    if os.path.isfile(generation_vs_error_file_path):
        os.remove(generation_vs_error_file_path)
    if os.path.isfile(neural_network_file_path):
        os.remove(neural_network_file_path)

    first_generation_population, successfully_retrieved_case_numbers = file_repository.get_population(
        generation_number=previous_generation_number)

    file_repository.population_path = presenter_output_path
    DataWriter.write_outputs(first_generation_population, successfully_retrieved_case_numbers,
                             file_repository, default_settings, generation_number=2, history=None)

    os.remove(summary_file)
    os.remove(generation_vs_error_file_path)
    file_repository.population_path = reax_output_dir_path
    neural_net_propagator = GeneticNeuralNetPropagator(default_settings, file_repository)
    model, history = neural_net_propagator.train_neural_net()
    file_repository.population_path = presenter_output_path
    DataWriter.write_outputs(first_generation_population, successfully_retrieved_case_numbers,
                             file_repository, default_settings, generation_number=2, history=history)




# Standard library
import os

# 3rd party packages
import pytest

# Local source


@pytest.fixture()
def training_set_dir_path():
    tests_dir = os.path.join(os.getcwd(), "tests")
    return os.path.join(tests_dir, "integration", "reference_training_set",
                        "ReaxFF_ZnO_Raymand_with_Sglass_control_all_bounds")


@pytest.fixture()
def reax_output_dir_path():
    tests_dir = os.path.join(os.getcwd(), "tests")
    return os.path.join(tests_dir, "integration", "reax_outputs")


@pytest.fixture()
def presenter_output_path():
    tests_dir = os.path.join(os.getcwd(), "tests")
    return os.path.join(tests_dir, "integration", "presenter")


@pytest.fixture()
def cli_output_path():
    tests_dir = os.path.join(os.getcwd(), "tests")
    return os.path.join(tests_dir, "integration", "cli")



#!/usr/bin/env python

# Standard library
from typing import Tuple
import random

# 3rd party packages


# Local source
from parametrization_clean.domain.crossover.strategy import ICrossoverStrategy
from parametrization_clean.domain.individual import Individual
from parametrization_clean.domain.root_individual import RootIndividual


class SinglePointCross(ICrossoverStrategy):

    @staticmethod
    def crossover(parent1: Individual, parent2: Individual, root_individual: RootIndividual,
                  **kwargs) -> Tuple[Individual, Individual]:
        """Execute single-point crossover. Cut the parameter vectors from two children at random positions
        and join to yield two new vectors (children).

        :param parent1: The first individual participating in crossover.
        :param parent2: The second individual participating in crossover.
        :return: Two new Children generated from the mating/crossover.
        """
        choice = random.randrange(0, len(parent1.params))
        child1 = Individual(parent1.params[:choice] + parent2.params[choice:], root_individual=root_individual)
        child2 = Individual(parent2.params[:choice] + parent1.params[choice:], root_individual=root_individual)
        return child1, child2



#!/usr/bin/env python

# Standard library
import random

# 3rd party packages


# Local source
from parametrization_clean.domain.mutation.strategy import IMutationStrategy
from parametrization_clean.domain.individual import Individual
from parametrization_clean.domain.root_individual import RootIndividual


class PolynomialMutate(IMutationStrategy):

    @staticmethod
    def mutation(parent: Individual, root_individual: RootIndividual, **kwargs) -> Individual:
        """Polynomial mutation adapted from Deb and Agrawal's paper.
        ADAPTED TO FUNCTION WITHOUT UPPER/LOWER BOUNDS FOR PARAMETERS.
        """
        eta = kwargs.get('polynomial_eta', 60)
        poly_degree = 1 / (1 + eta)
        new_params = []
        for param in parent.params:
            u = random.random()
            if u <= 0.5:
                delta_l = (2 * u) ** poly_degree - 1
                param = param + delta_l * param
            else:
                delta_r = 1 - (2 * (1 - u)) ** poly_degree
                param = param + delta_r * param
            new_params.append(param)
        return Individual(new_params, root_individual=root_individual)




# Standard library

# 3rd party packages

# Local source
from parametrization_clean.domain.selection.tournament import TournamentSelect
from parametrization_clean.domain.mutation.gauss import GaussianMutate
from parametrization_clean.domain.crossover.double_pareto import DoubleParetoCross
from parametrization_clean.domain.adaptation.xiao import XiaoAdapt
from parametrization_clean.domain.cost.reax_error import ReaxError
from parametrization_clean.domain.mutation.nakata import NakataMutate
from parametrization_clean.infrastructure.config.default import DefaultSettings


def test_all_settings_init():
    default_settings = DefaultSettings()

    assert default_settings.strategy_settings.selection_strategy == TournamentSelect
    assert default_settings.strategy_settings.error_strategy == ReaxError
    assert default_settings.strategy_settings.adaptation_strategy == XiaoAdapt
    assert default_settings.strategy_settings.crossover_strategy == DoubleParetoCross
    assert default_settings.strategy_settings.mutation_strategy == GaussianMutate
    assert default_settings.strategy_settings.initialization_strategy == NakataMutate

    assert default_settings.ga_settings.population_size == 30
    assert default_settings.ga_settings.mutation_rate == 0.2
    assert default_settings.ga_settings.crossover_rate == 0.8
    assert default_settings.ga_settings.use_elitism
    assert not default_settings.ga_settings.use_adaptation

    assert default_settings.mutation_settings.gauss_std == [0.01, 0.1]
    assert default_settings.mutation_settings.gauss_frac == [0.5, 0.5]
    assert default_settings.mutation_settings.nakata_rand_lower == -1.0
    assert default_settings.mutation_settings.nakata_rand_higher == 1.0
    assert default_settings.mutation_settings.nakata_scale == 0.1
    assert default_settings.mutation_settings.polynomial_eta == 60
    assert not default_settings.mutation_settings.param_bounds

    assert default_settings.crossover_settings.dpx_alpha == 10
    assert default_settings.crossover_settings.dpx_beta == 1

    assert default_settings.selection_settings.tournament_size == 2

    assert default_settings.adaptation_settings.srinivas_k1 == 1.0
    assert default_settings.adaptation_settings.srinivas_k2 == 0.5
    assert default_settings.adaptation_settings.srinivas_k3 == 1.0
    assert default_settings.adaptation_settings.srinivas_k4 == 0.5
    assert default_settings.adaptation_settings.srinivas_default_mutation_rate == 0.01
    assert default_settings.adaptation_settings.xiao_min_crossover_rate == 0.8 * 0.8
    assert default_settings.adaptation_settings.xiao_min_mutation_rate == 0.2 * 0.8
    assert default_settings.adaptation_settings.xiao_scale == 40

    assert default_settings.neural_net_settings.verbosity == 2
    assert default_settings.neural_net_settings.train_fraction == 0.8
    assert default_settings.neural_net_settings.num_epochs == 10000
    assert default_settings.neural_net_settings.num_populations_to_train_on == 1
    assert default_settings.neural_net_settings.num_nested_ga_iterations == 1



#!/usr/bin/env python

"""Module with interface for selection operators to be performed on Individuals.
New selection strategies can be added as classes, so long as they implement the abstraction presented here.
"""

# Standard library
import abc
from typing import List

# 3rd party packages

# Local source
from parametrization_clean.domain.individual import Individual


class ISelectionStrategy(metaclass=abc.ABCMeta):

    @staticmethod
    @abc.abstractmethod
    def selection(population: List[Individual], **kwargs) -> Individual:
        raise NotImplementedError



#!/usr/bin/env python

"""Factory for crossover algorithms allowed for usage."""

# Standard library

# 3rd party packages

# Local source
from parametrization_clean.domain.crossover.strategy import ICrossoverStrategy
from parametrization_clean.domain.crossover.single_point import SinglePointCross
from parametrization_clean.domain.crossover.two_point import TwoPointCross
from parametrization_clean.domain.crossover.uniform import UniformCross
from parametrization_clean.domain.crossover.double_pareto import DoubleParetoCross


class CrossoverFactory:
    """Factory class for creating crossover algorithm executor - RegistryHolder design pattern.
    Classes that implement ICrossoverStrategy can be registered and utilized through this factory's registry.
    """

    REGISTRY = {}
    """Internal registry for available crossover methods. Users can specify from one of the
    `algorithm_name` strings available in the dictionary, mapping `algorithm_name` to the corresponding class
    implementing that algorithm.
    For example, "single_point" maps to the single point crossover algorithm (for real-valued GAs);
    users can specify the `crossover_strategy` in the user config.json file to use this algorithm.
    """

    @classmethod
    def register(cls, algorithm_name: str, crossover_class):
        """Register a crossover strategy with a string key. Useful for abstraction and dynamic retrieval
        of different algorithms in configuration file. Using this factory, one can easily implement a crossover
        algorithm (ex. MyCrossoverClass) that follows ICrossoverStrategy, then use
        "CrossoverFactory.register('my_crossover_class_name')"
        to generate a corresponding string reference for that crossover strategy.

        Parameters
        ----------
        algorithm_name: str
            Name that one wishes to assign to the designated `crossover_class`/algorithm.
        crossover_class
            Class that one wishes to associate/register with `algorithm_name`.
        Returns
        -------
        crossover_class
            Same as the `crossover_class` input parameter.
        """
        cls.REGISTRY[algorithm_name] = crossover_class
        return crossover_class

    @classmethod
    def create_executor(cls, algorithm_name: str) -> ICrossoverStrategy:
        return cls.REGISTRY[algorithm_name]


CrossoverFactory.register('single_point', SinglePointCross)
CrossoverFactory.register('two_point', TwoPointCross)
CrossoverFactory.register('uniform', UniformCross)
CrossoverFactory.register('double_pareto', DoubleParetoCross)




# Standard library

# 3rd party packages

# Local source
from parametrization_clean.domain.selection.factory import SelectionFactory
from parametrization_clean.domain.selection.tournament import TournamentSelect


def test_get_tournament():
    assert SelectionFactory.create_executor('tournament') == TournamentSelect



#!/usr/bin/env python

# Standard library
import random

# 3rd party packages


# Local source
from parametrization_clean.domain.mutation.strategy import IMutationStrategy
from parametrization_clean.domain.individual import Individual
from parametrization_clean.domain.root_individual import RootIndividual


class CentralUniformMutate(IMutationStrategy):

    @staticmethod
    def mutation(parent: Individual, root_individual: RootIndividual, **kwargs) -> Individual:
        """Inspired by Monte Carlo/GA guidelines for ReaxFF paper.
        Use a random number (determined by uniform distribution) in the central segment for each parameter range.
        So, if a parameter has a range of [p_min, p_max], a uniform random number will be generated in the bounds
        [p_min + (p_max - p_min)/4, p_max - (p_max - p_min)/4].
        Mutates all parameters (doesn't consider `FRAC_PARAMS_MUTATE`).
        """
        param_bounds = kwargs['param_bounds']
        new_params = []
        for (lower_bound, upper_bound) in param_bounds:
            delta = (upper_bound - lower_bound) / 4
            new_param = random.uniform(lower_bound + delta, upper_bound - delta)
            new_params.append(new_param)
        return Individual(new_params, root_individual=root_individual)



#!/usr/bin/env python

# Standard library
import random

# 3rd party packages


# Local source
from parametrization_clean.domain.mutation.strategy import IMutationStrategy
from parametrization_clean.domain.individual import Individual
from parametrization_clean.domain.root_individual import RootIndividual


class GaussianMutate(IMutationStrategy):

    @staticmethod
    def mutation(parent: Individual, root_individual: RootIndividual, **kwargs) -> Individual:
        """UNCONSTRAINED Gaussian mutation.
        Upper and lower bounds are not required for this version.
        Allows usage of multiple scaling factors for the normal distribution for mutation.
        Each scaling factor needs a corresponding probability (gauss_frac) of using that given factor.
        """
        stds = kwargs.get('gauss_std', [0.1])
        probabilities = kwargs.get('gauss_frac', [1.0])
        u = random.uniform(0, 1)
        cumulative_probability = 0
        for std, probability in zip(stds, probabilities):
            if cumulative_probability < u <= (cumulative_probability + probability):
                new_params = [param + param * random.gauss(0, std) for param in parent.params]
                break
            cumulative_probability += probability
        return Individual(new_params, root_individual=root_individual)



#!/usr/bin/env python

"""Module with immutable structure to store reference training set information.
Potentially includes DFT energies, weights, parameter boundaries (low, high)."""

# Standard library
from typing import List, Dict


# 3rd party packages

# Local source
from parametrization_clean.domain.utils.helpers import get_param


class Borg(object):
    """Avoid Singleton anti-pattern, but ensure shared state (if many instances are created)."""
    _shared_state = {}

    def __init__(self):
        self.__dict__ = self._shared_state


class RootIndividual(Borg):

    def __init__(self, dft_energies: List[float], weights: List[float],
                 root_ffield: Dict[int, List], param_keys: List[List[int]]):
        """Root individual to store data from reference training set that needs to be stored only once.
        For example, the weights of the error, DFT energies, parameter bounds only need to be stored once.
        Input structures are converted to tuples to prevent mutation.

        Parameters
        ----------
        dft_energies: List[float]
            Stored as a tuple containing DFT energies; used to compute error.
        weights: List[float]
            Stored as a tuple containing weights, used to compute error.
        root_ffield: Dict[int, List]
            Dictionary mapping ReaxFF force field section number to parameters contained in that section.
        param_keys: List[List[int]]
            List of keys mapping to values in the ffield object.
        """
        super().__init__()
        self.dft_energies = tuple(dft_energies)
        self.weights = tuple(weights)
        self.root_ffield = root_ffield
        self.param_keys = param_keys

        # Root parameters from the reference training set
        self.root_params = self.extract_params()

    def extract_params(self) -> List[float]:
        return [get_param(key, self.root_ffield) for key in self.param_keys]


class FirstGenerationRootIndividual(RootIndividual):

    def __init__(self, root_ffield: Dict[int, List], param_keys: List[List[int]]):
        """Root individual for the first generation. Does NOT require DFT energies or weights, as they are not
        necessary for population initialization.

        Parameters
        ----------
        root_ffield: Dict[int, List]
            Dictionary mapping ReaxFF force field section number to parameters contained in that section.
        param_keys: List[List[int]]
            List of keys mapping to values in the ffield object.
        """
        super().__init__(dft_energies=[], weights=[], root_ffield=root_ffield, param_keys=param_keys)



#!/usr/bin/env python

# Standard library
from typing import Tuple, List

# 3rd party packages


# Local source
from parametrization_clean.domain.adaptation.strategy import IAdaptationStrategy


class SrinivasAdapt(IAdaptationStrategy):

    @staticmethod
    def adaptation(average_cost: float, minimum_cost: float, parent_costs: Tuple[float, float], **kwargs) \
            -> Tuple[float, List[float]]:
        """Execute Srinivas adaptive algorithm. See Srinivas' 1994 paper on "Adaptive probabilities of crossover
        and mutation in genetic algorithms" for more information.

        Parameters
        ----------
        average_cost: float
            Measure of central tendency of cost of population. Can also be the median.
        minimum_cost: float
            Lowest cost of population of Individuals.
        parent_costs: Tuple[float, float]
            Tuple containing costs associated with two parents.
        kwargs: Dict
            Retrieves parameters associated with executing this algorithm from the passed optional dict; otherwise,
            defaults are used.

        Returns
        -------
        cross_rate: float
            Updated crossover rate.
        mutation_rates: List[float, float]
            Updated mutation rates for each of the two children about to be generated from the two parents.
        """
        k1 = kwargs.get('srinivas_k1', 1.0)
        k2 = kwargs.get('srinivas_k2', 0.5)
        k3 = kwargs.get('srinivas_k3', 1.0)
        k4 = kwargs.get('srinivas_k4', 0.5)
        default_mutation_rate = kwargs.get('srinivas_default_mutation_rate', 0.005)

        spread = average_cost - minimum_cost

        better_parent_cost = min(parent_costs)
        if better_parent_cost <= average_cost:
            cross_rate = k1 * (better_parent_cost - minimum_cost) / spread
        else:
            cross_rate = k3

        mutation_rates = []
        for cost in parent_costs:
            if cost <= average_cost:
                new_mutation_rate = max(default_mutation_rate, k2 * (cost - minimum_cost) / spread)
                mutation_rates.append(new_mutation_rate)
            else:
                mutation_rates.append(k4)

        return cross_rate, mutation_rates




# Standard library
from unittest.mock import patch

# 3rd party packages
import pytest

# Local source
from parametrization_clean.domain.selection.tournament import TournamentSelect


@patch('random.sample')
@pytest.mark.usefixtures('get_individuals')
def test_tournament(sample_mock, get_individuals):
    get_individuals[0].cost = 5142
    get_individuals[1].cost = 4387
    get_individuals[2].cost = 5789
    get_individuals[3].cost = 4698
    sample_mock.return_value = get_individuals[0], get_individuals[2]
    assert TournamentSelect.selection(get_individuals) == get_individuals[0]
    sample_mock.return_value = get_individuals
    assert TournamentSelect.selection(get_individuals, tournament_size=4) == get_individuals[1]



#!/usr/bin/env python

"""
Module that contains interface for crossover methods to be used for an Individual.
New crossover strategies can be added as classes, so long as they implement the abstraction presented here.

__author__ = "Chad Daksha"
"""

# Standard library
import abc
from typing import Tuple

# 3rd party packages


# Local source
from parametrization_clean.domain.individual import Individual
from parametrization_clean.domain.root_individual import RootIndividual


class ICrossoverStrategy(metaclass=abc.ABCMeta):

    @staticmethod
    @abc.abstractmethod
    def crossover(parent1: Individual, parent2: Individual, root_individual: RootIndividual,
                  **kwargs) -> Tuple[Individual, Individual]:
        raise NotImplementedError



#!/usr/bin/env python

"""Module that contains interface for settings repository, used to get configuration settings.
All settings are subdivided into different categories based on groupings/relevance.
"""

# Standard library
import abc
from typing import List

# 3rd party packages

# Local source
from parametrization_clean.domain.selection.strategy import ISelectionStrategy
from parametrization_clean.domain.crossover.strategy import ICrossoverStrategy
from parametrization_clean.domain.mutation.strategy import IMutationStrategy
from parametrization_clean.domain.adaptation.strategy import IAdaptationStrategy
from parametrization_clean.domain.cost.strategy import IErrorStrategy


class IStrategySettings(abc.ABC):

    @abc.abstractmethod
    def __init__(self):
        self.selection_strategy: ISelectionStrategy = NotImplemented
        self.mutation_strategy: IMutationStrategy = NotImplemented
        self.crossover_strategy: ICrossoverStrategy = NotImplemented
        self.adaptation_strategy: IAdaptationStrategy = NotImplemented
        self.error_strategy: IErrorStrategy = NotImplemented
        self.initialization_strategy: IMutationStrategy = NotImplemented


class IGeneticAlgorithmSettings:

    @abc.abstractmethod
    def __init__(self):
        self.population_size: int = NotImplemented
        self.mutation_rate: float = NotImplemented
        self.crossover_rate: float = NotImplemented
        self.use_elitism: bool = NotImplemented
        self.use_adaptation: bool = NotImplemented
        self.use_neural_network: bool = NotImplemented


class IMutationSettings(abc.ABC):

    @abc.abstractmethod
    def __init__(self):
        self.gauss_std: List[float] = NotImplemented
        self.gauss_frac: List[float] = NotImplemented

        self.nakata_rand_lower: float = NotImplemented
        self.nakata_rand_higher: float = NotImplemented
        self.nakata_scale: float = NotImplemented

        self.polynomial_eta: float = NotImplemented

        self.param_bounds: List[List[float]] = NotImplemented


class ICrossoverSettings(abc.ABC):

    @abc.abstractmethod
    def __init__(self):
        self.dpx_alpha: float = NotImplemented
        self.dpx_beta: float = NotImplemented


class ISelectionSettings(abc.ABC):

    @abc.abstractmethod
    def __init__(self):
        self.tournament_size: int = NotImplemented


class IAdaptationSettings(abc.ABC):

    @abc.abstractmethod
    def __init__(self):
        self.srinivas_k1: float = NotImplemented
        self.srinivas_k2: float = NotImplemented
        self.srinivas_k3: float = NotImplemented
        self.srinivas_k4: float = NotImplemented
        self.srinivas_default_mutation_rate: float = NotImplemented

        self.xiao_min_crossover_rate: float = NotImplemented
        self.xiao_min_mutation_rate: float = NotImplemented
        self.xiao_scale: float = NotImplemented


class INeuralNetSettings(abc.ABC):

    @abc.abstractmethod
    def __init__(self):
        self.verbosity: int = NotImplemented
        self.train_fraction: float = NotImplemented
        self.num_epochs: int = NotImplemented
        self.num_populations_to_train_on: int = NotImplemented
        self.num_nested_ga_iterations: int = NotImplemented
        self.minimum_validation_r_squared: float = NotImplemented


class IAllSettings(abc.ABC):

    @abc.abstractmethod
    def __init__(self):
        self.strategy_settings: IStrategySettings = NotImplemented
        self.ga_settings: IGeneticAlgorithmSettings = NotImplemented
        self.mutation_settings: IMutationSettings = NotImplemented
        self.crossover_settings: ICrossoverSettings = NotImplemented
        self.selection_settings: ISelectionSettings = NotImplemented
        self.neural_net_settings: INeuralNetSettings = NotImplemented
